import os

import numpy as np
from dpdata.abacus.scf import get_cell, get_coords, get_nele_from_stru

from . import vasp_utils
from dflow.python import upload_packages
from dargs.dargs import Argument
upload_packages.append(__file__)

bohr2ang = 0.52917721067


def make_abacus_scf_kpt(fp_params):
    # Make KPT file for abacus pw scf calculation.
    # KPT file is the file containing k points infomation in ABACUS scf calculation.
    k_points = [1, 1, 1, 0, 0, 0]
    if "k_points" in fp_params:
        k_points = fp_params["k_points"]
        if len(k_points) != 6:
            raise RuntimeError(
                "k_points has to be a list containig 6 integers specifying MP k points generation."
            )
    ret = "K_POINTS\n0\nGamma\n"
    for i in range(6):
        ret += str(k_points[i]) + " "
    return ret


def make_abacus_scf_input(fp_params):
    # Make INPUT file for abacus pw scf calculation.
    ret = "INPUT_PARAMETERS\n"
    ret += "calculation scf\n"
    exam_float = Argument("base", float)
    exam_int = Argument("base", int)
    for key in fp_params:
        if key == "ecutwfc":
            exam_float.check({"ecutwfc": fp_params["ecutwfc"]})
            #fp_params["ecutwfc"] = float(fp_params["ecutwfc"])
            assert fp_params["ecutwfc"] >= 0, "'ecutwfc' should be non-negative."
            ret += "ecutwfc %f\n" % fp_params["ecutwfc"]
        elif key == "kspacing":
            exam_float.check({"kspacing": fp_params["kspacing"]})
            #fp_params["kspacing"] = float(fp_params["kspacing"])
            assert fp_params["kspacing"] >= 0, "'kspacing' should be non-negative."
            ret += "kspacing %f\n" % fp_params["kspacing"]
        elif key == "scf_thr":
            exam_float.check({"scf_thr": fp_params["scf_thr"]})
            #fp_params["scf_thr"] = float(fp_params["scf_thr"])
            ret += "scf_thr %e\n" % fp_params["scf_thr"]
        elif key == "scf_nmax":
            exam_int.check({"scf_nmax": fp_params["scf_nmax"]})
            #fp_params["scf_nmax"] = int(fp_params["scf_nmax"])
            assert fp_params["scf_nmax"] >= 0, "'scf_nmax' should be a positive integer."
            ret += "scf_nmax %d\n" % fp_params["scf_nmax"]
        elif key == "basis_type":
            assert fp_params["basis_type"] in [
                "pw",
                "lcao",
                "lcao_in_pw",
            ], "'basis_type' must in 'pw', 'lcao' or 'lcao_in_pw'."
            ret += "basis_type %s\n" % fp_params["basis_type"]
        elif key == "dft_functional":
            ret += "dft_functional %s\n" % fp_params["dft_functional"]
        elif key == "gamma_only":
            if type(fp_params["gamma_only"]) == str:
                fp_params["gamma_only"] = int(eval(fp_params["gamma_only"]))
            assert (
                fp_params["gamma_only"] == 0 or fp_params["gamma_only"] == 1
            ), "'gamma_only' should be either 0 or 1."
            ret += "gamma_only %d\n" % fp_params["gamma_only"]
        elif key == "mixing_type":
            assert fp_params["mixing_type"] in [
                "plain",
                "kerker",
                "pulay",
                "pulay-kerker",
                "broyden",
            ]
            ret += "mixing_type %s\n" % fp_params["mixing_type"]
        elif key == "mixing_beta":
            exam_float.check({"mixing_beta": fp_params["mixing_beta"]})
            #fp_params["mixing_beta"] = float(fp_params["mixing_beta"])
            assert (
                fp_params["mixing_beta"] >= 0 and fp_params["mixing_beta"] < 1
            ), "'mixing_beta' should between 0 and 1."
            ret += "mixing_beta %f\n" % fp_params["mixing_beta"]
        elif key == "symmetry":
            if type(fp_params["symmetry"]) == str:
                fp_params["symmetry"] = int(eval(fp_params["symmetry"]))
            assert (
                fp_params["symmetry"] == 0 or fp_params["symmetry"] == 1
            ), "'symmetry' should be either 0 or 1."
            ret += "symmetry %d\n" % fp_params["symmetry"]
        elif key == "nbands":
            exam_int.check({"nbands": fp_params["nbands"]})
            #fp_params["nbands"] = int(fp_params["nbands"])
            assert (
                fp_params["nbands"] > 0 and type(fp_params["nbands"]) == int
            ), "'nbands' should be a positive integer."
            ret += "nbands %d\n" % fp_params["nbands"]
        elif key == "nspin":
            exam_int.check({"nspin": fp_params["nspin"]})
            #fp_params["nspin"] = int(fp_params["nspin"])
            assert (
                fp_params["nspin"] == 1
                or fp_params["nspin"] == 2
                or fp_params["nspin"] == 4
            ), "'nspin' can anly take 1, 2 or 4"
            ret += "nspin %d\n" % fp_params["nspin"]
        elif key == "ks_solver":
            assert fp_params["ks_solver"] in [
                "cg",
                "dav",
                "lapack",
                "genelpa",
                "hpseps",
                "scalapack_gvx",
            ], "'ks_sover' should in 'cgx', 'dav', 'lapack', 'genelpa', 'hpseps', 'scalapack_gvx'."
            ret += "ks_solver %s\n" % fp_params["ks_solver"]
        elif key == "smearing_method":
            assert fp_params["smearing_method"] in [
                "gauss",
                "gaussian",
                "fd",
                "fixed",
                "mp",
                "mp2",
                "mv",
            ], "'smearing_method' should in 'gauss', 'gaussian', 'fd', 'fixed', 'mp', 'mp2', 'mv'. "
            ret += "smearing_method %s\n" % fp_params["smearing_method"]
        elif key == "smearing_sigma":
            exam_float.check({"smearing_sigma": fp_params["smearing_sigma"]})
            #fp_params["smearing_sigma"] = float(fp_params["smearing_sigma"])
            assert (
                fp_params["smearing_sigma"] >= 0
            ), "'smearing_sigma' should be non-negative."
            ret += "smearing_sigma %f\n" % fp_params["smearing_sigma"]
        elif key == "cal_force":
            if type(fp_params["cal_force"]) == str:
                fp_params["cal_force"] = int(eval(fp_params["cal_force"]))
            assert (
                fp_params["cal_force"] == 0 or fp_params["cal_force"] == 1
            ), "'cal_force' should be either 0 or 1."
            ret += "cal_force %d\n" % fp_params["cal_force"]
        elif key == "cal_stress":
            if type(fp_params["cal_stress"]) == str:
                fp_params["cal_stress"] = int(eval(fp_params["cal_stress"]))
            assert (
                fp_params["cal_stress"] == 0 or fp_params["cal_stress"] == 1
            ), "'cal_stress' should be either 0 or 1."
            ret += "cal_stress %d\n" % fp_params["cal_stress"]
        # paras for deepks
        elif key == "deepks_out_labels":
            if type(fp_params["deepks_out_labels"]) == str:
                fp_params["deepks_out_labels"] = int(
                    eval(fp_params["deepks_out_labels"])
                )
            assert (
                fp_params["deepks_out_labels"] == 0
                or fp_params["deepks_out_labels"] == 1
            ), "'deepks_out_labels' should be either 0 or 1."
            ret += "deepks_out_labels %d\n" % fp_params["deepks_out_labels"]
        elif key == "deepks_descriptor_lmax":
            exam_int.check({"deepks_descriptor_lmax": fp_params["deepks_descriptor_lmax"]})
            #fp_params["deepks_descriptor_lmax"] = int(
            #    fp_params["deepks_descriptor_lmax"]
            #)
            assert (
                fp_params["deepks_descriptor_lmax"] >= 0
            ), "'deepks_descriptor_lmax' should be  a positive integer."
            ret += "deepks_descriptor_lmax %d\n" % fp_params["deepks_descriptor_lmax"]
        elif key == "deepks_scf":
            if type(fp_params["deepks_scf"]) == str:
                fp_params["deepks_scf"] = int(eval(fp_params["deepks_scf"]))
            assert (
                fp_params["deepks_scf"] == 0 or fp_params["deepks_scf"] == 1
            ), "'deepks_scf' should be either 0 or 1."
            ret += "deepks_scf %d\n" % fp_params["deepks_scf"]
        elif key == "deepks_model":
            ret += "deepks_model %s\n" % fp_params["deepks_model"]
        elif key[0] == "_":
            pass
        elif key == "calculation":
            pass
        else:
            ret += "%s %s\n" % (key, str(fp_params[key]))
    return ret


def make_abacus_scf_stru(
    sys_data,
    fp_pp_files,
    fp_orb_files=None,
    fp_dpks_descriptor=None,
    fp_params=None,
    type_map=None,
):
    atom_names = sys_data["atom_names"]
    atom_numbs = sys_data["atom_numbs"]
    if type_map == None:
        type_map = atom_names

    assert len(atom_names) == len(atom_numbs), "Please check the name of atoms. "
    cell = sys_data["cells"].reshape([3, 3])
    coord = sys_data["coords"].reshape([sum(atom_numbs), 3])
    # volume = np.linalg.det(cell)
    # lattice_const = np.power(volume, 1/3)

    ret = "ATOMIC_SPECIES\n"
    for iatom in range(len(atom_names)):
        assert atom_names[iatom] in type_map, (
            "element %s is not defined in type_map" % atom_names[iatom]
        )
        idx = type_map.index(atom_names[iatom])
        if "atom_masses" not in sys_data:
            ret += atom_names[iatom] + " 1.00 " + fp_pp_files[idx] + "\n"
        else:
            ret += (
                atom_names[iatom]
                + " %.3f " % sys_data["atom_masses"][iatom]
                + fp_pp_files[idx]
                + "\n"
            )

    if fp_params is not None and "lattice_constant" in fp_params:
        ret += "\nLATTICE_CONSTANT\n"
        ret += (
            str(fp_params["lattice_constant"]) + "\n\n"
        )  # in Bohr, in this way coord and cell are in Angstrom
    else:
        ret += "\nLATTICE_CONSTANT\n"
        ret += str(1 / bohr2ang) + "\n\n"

    ret += "LATTICE_VECTORS\n"
    for ix in range(3):
        for iy in range(3):
            ret += str(cell[ix][iy]) + " "
        ret += "\n"
    ret += "\n"

    ret += "ATOMIC_POSITIONS\n"
    ret += "Cartesian    # Cartesian(Unit is LATTICE_CONSTANT)\n"
    # ret += "\n"
    natom_tot = 0
    for iele in range(len(atom_names)):
        ret += atom_names[iele] + "\n"
        ret += "0.0\n"
        ret += str(atom_numbs[iele]) + "\n"
        for iatom in range(atom_numbs[iele]):
            ret += "%.12f %.12f %.12f %d %d %d\n" % (
                coord[natom_tot, 0],
                coord[natom_tot, 1],
                coord[natom_tot, 2],
                1,
                1,
                1,
            )
            natom_tot += 1
    assert natom_tot == sum(atom_numbs)

    if fp_orb_files is not None:
        ret += "\nNUMERICAL_ORBITAL\n"
        assert len(fp_orb_files) == len(type_map)
        for iatom in range(len(atom_names)):
            idx = type_map.index(atom_names[iatom])
            ret += fp_orb_files[idx] + "\n"

    if fp_dpks_descriptor is not None:
        ret += "\nNUMERICAL_DESCRIPTOR\n"
        ret += "%s\n" % fp_dpks_descriptor

    return ret


def get_abacus_input_parameters(INPUT):
    with open(INPUT) as fp:
        inlines = fp.read().split("\n")
    input_parameters = {}
    for line in inlines:
        if line.split() == [] or len(line.split()) < 2 or line[0] in ["#"]:
            continue
        parameter_name = line.split()[0]
        parameter_value = line.split()[1]
        input_parameters[parameter_name] = parameter_value
    fp.close()
    return input_parameters


def get_mass_from_STRU(geometry_inlines, atom_names):
    nele = get_nele_from_stru(geometry_inlines)
    assert nele > 0
    mass_list = [0 for i in atom_names]
    pp_file_list = [i for i in atom_names]
    for iline, line in enumerate(geometry_inlines):
        if line.split() == []:
            continue
        if "ATOMIC_SPECIES" == line.split()[0]:
            for iele1 in range(1, 1 + nele):
                for iele2 in range(nele):
                    if geometry_inlines[iline + iele1].split()[0] == atom_names[iele2]:
                        mass_list[iele2] = float(
                            geometry_inlines[iline + iele1].split()[1]
                        )
                        pp_file_list[iele2] = geometry_inlines[iline + iele1].split()[2]
    for iele in range(len(mass_list)):
        assert mass_list[iele] > 0
    return mass_list, pp_file_list


def get_natoms_from_stru(geometry_inlines):
    key_words_list = [
        "ATOMIC_SPECIES",
        "NUMERICAL_ORBITAL",
        "LATTICE_CONSTANT",
        "LATTICE_VECTORS",
        "ATOMIC_POSITIONS",
        "NUMERICAL_DESCRIPTOR",
    ]
    keyword_sequence = []
    keyword_line_index = []
    atom_names = []
    atom_numbs = []
    tmp_line = []
    for i in geometry_inlines:
        if i.strip() != "":
            tmp_line.append(i)
    for iline, line in enumerate(tmp_line):
        if line.split() == []:
            continue
        have_key_word = False
        for keyword in key_words_list:
            if keyword in line and keyword == line.split()[0]:
                keyword_sequence.append(keyword)
                keyword_line_index.append(iline)
    assert len(keyword_line_index) == len(keyword_sequence)
    assert len(keyword_sequence) > 0
    keyword_line_index.append(len(tmp_line))
    for idx, keyword in enumerate(keyword_sequence):
        if keyword == "ATOMIC_POSITIONS":
            iline = keyword_line_index[idx] + 2
            while iline < keyword_line_index[idx + 1] - 1:
                atom_names.append(tmp_line[iline].split()[0])
                atom_numbs.append(int(tmp_line[iline + 2].split()[0]))
                iline += 3 + atom_numbs[-1]
    return atom_names, atom_numbs


def get_additional_from_STRU(geometry_inlines, nele):
    dpks_descriptor_kw = "NUMERICAL_DESCRIPTOR"
    orb_file_kw = "NUMERICAL_ORBITAL"
    dpks_descriptor = None
    orb_file = None
    for iline in range(len(geometry_inlines)):
        if len(geometry_inlines[iline]) > 0:
            if orb_file_kw == geometry_inlines[iline].split()[0]:
                orb_file = []
                for iele in range(nele):
                    orb_file.append(geometry_inlines[iline + iele + 1].strip())
            if dpks_descriptor_kw == geometry_inlines[iline].split()[0]:
                dpks_descriptor = geometry_inlines[iline + 1].strip()
    return orb_file, dpks_descriptor


def get_abacus_STRU(STRU, INPUT=None, n_ele=None):
    # read in geometry from STRU file. n_ele is the number of elements.
    # Either n_ele or INPUT should be provided.
    with open(STRU, "r") as fp:
        geometry_inlines = fp.read().split("\n")
    for iline, line in enumerate(geometry_inlines):
        if line.split() == [] or len(line) == 0:
            del geometry_inlines[iline]
    geometry_inlines.append("")
    celldm, cell = get_cell(geometry_inlines)
    atom_names, natoms, types, coords = get_coords(celldm, cell, geometry_inlines)
    masses, pp_files = get_mass_from_STRU(geometry_inlines, atom_names)
    orb_files, dpks_descriptor = get_additional_from_STRU(geometry_inlines, len(masses))
    data = {}
    data["atom_names"] = atom_names
    data["atom_numbs"] = natoms
    data["atom_types"] = types
    data["cells"] = cell
    data["coords"] = coords
    data[
        "atom_masses"
    ] = masses  # Notice that this key is not defined in dpdata system.
    data["pp_files"] = pp_files
    data["orb_files"] = orb_files
    data["dpks_descriptor"] = dpks_descriptor
    return data


def make_supercell_abacus(from_struct, super_cell):
    if "types" in from_struct:
        from_struct["types"] = (
            from_struct["types"] * super_cell[0] * super_cell[1] * super_cell[2]
        )
    for ix in range(super_cell[0]):
        for iy in range(super_cell[1]):
            for iz in range(super_cell[2]):
                if ix == 0 and iy == 0 and iz == 0:
                    continue
                for ia in range(sum(from_struct["atom_numbs"])):
                    coord = (
                        from_struct["coords"][ia]
                        + from_struct["cells"][0] * ix
                        + from_struct["cells"][1] * iy
                        + from_struct["cells"][2] * iz
                    )
                    from_struct["coords"] = np.vstack([from_struct["coords"], coord])
    from_struct["atom_numbs"] = [
        i * super_cell[0] * super_cell[1] * super_cell[2]
        for i in from_struct["atom_numbs"]
    ]
    from_struct["cells"][0] *= super_cell[0]
    from_struct["cells"][1] *= super_cell[1]
    from_struct["cells"][2] *= super_cell[2]
    return from_struct


def make_kspacing_kpoints_stru(stru, kspacing):
    # adapted from dpgen.autotest.lib.vasp.make_kspacing_kpoints
    if type(kspacing) is not list:
        kspacing = [kspacing, kspacing, kspacing]
    box = stru["cells"]
    rbox = vasp_utils.reciprocal_box(box)
    kpoints = [
        max(1, (np.ceil(2 * np.pi * np.linalg.norm(ii) / ks).astype(int)))
        for ii, ks in zip(rbox, kspacing)
    ]
    kpoints += [0, 0, 0]
    return kpoints


if __name__ == "__main__":
    fp_params = {"k_points": [1, 1, 1, 0, 0, 0]}
    ret = make_abacus_scf_kpt(fp_params)
    print(ret)



from apex.core.calculator.ABACUS import ABACUS
from apex.core.calculator.Lammps import Lammps
from apex.core.calculator.VASP import VASP
from dflow.python import upload_packages
upload_packages.append(__file__)

LAMMPS_TYPE = ["deepmd", "meam", "eam_fs", "eam_alloy"]

def make_calculator(inter_parameter, path_to_poscar):
    """
    Make an instance of Task
    """
    inter_type = inter_parameter["type"]
    if inter_type == "vasp":
        return VASP(inter_parameter, path_to_poscar)
    elif inter_type == "abacus":
        return ABACUS(inter_parameter, path_to_poscar)
    elif inter_type in LAMMPS_TYPE:
        return Lammps(inter_parameter, path_to_poscar)
    #    if inter_type == 'siesta':
    #        return Siesta(inter_parameter, path_to_poscar)
    #        pass
    else:
        raise RuntimeError(f"unsupported interaction {inter_type}")




import glob
import json
import logging
import os
import shutil
import re
import subprocess
import dpdata
from pathlib import Path
from pymatgen.core.structure import Structure

from apex.core.structure import StructureInfo
from apex.core.calculator.calculator import LAMMPS_TYPE
from apex.core.calculator.lib import abacus_utils
from apex.core.calculator.lib import vasp_utils
from apex.core.property.Property import Property
from apex.core.refine import make_refine
from apex.core.reproduce import make_repro, post_repro
from dflow.python import upload_packages
upload_packages.append(__file__)


class Phonon(Property):
    def __init__(self, parameter, inter_param=None):
        parameter["reproduce"] = parameter.get("reproduce", False)
        self.reprod = parameter["reproduce"]
        if not self.reprod:
            if not ("init_from_suffix" in parameter and "output_suffix" in parameter):
                self.primitive = parameter.get('primitive', False)
                self.approach = parameter.get('approach', 'linear')
                self.supercell_size = parameter.get('supercell_size', [2, 2, 2])
                self.MESH = parameter.get('MESH', None)
                self.PRIMITIVE_AXES = parameter.get('PRIMITIVE_AXES', None)
                self.BAND = parameter.get('BAND', None)
                self.BAND_POINTS = parameter.get('BAND_POINTS', None)
                self.BAND_CONNECTION = parameter.get('BAND_CONNECTION', True)
            parameter["cal_type"] = parameter.get("cal_type", "relaxation")
            self.cal_type = parameter["cal_type"]
            default_cal_setting = {
                "relax_pos": True,
                "relax_shape": False,
                "relax_vol": False,
            }
            if "cal_setting" not in parameter:
                parameter["cal_setting"] = default_cal_setting
            else:
                if "relax_pos" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_pos"] = default_cal_setting[
                        "relax_pos"
                    ]
                if "relax_shape" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_shape"] = default_cal_setting[
                        "relax_shape"
                    ]
                if "relax_vol" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_vol"] = default_cal_setting[
                        "relax_vol"
                    ]
            self.cal_setting = parameter["cal_setting"]
        else:
            parameter["cal_type"] = "static"
            self.cal_type = parameter["cal_type"]
            default_cal_setting = {
                "relax_pos": False,
                "relax_shape": False,
                "relax_vol": False,
            }
            if "cal_setting" not in parameter:
                parameter["cal_setting"] = default_cal_setting
            else:
                if "relax_pos" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_pos"] = default_cal_setting[
                        "relax_pos"
                    ]
                if "relax_shape" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_shape"] = default_cal_setting[
                        "relax_shape"
                    ]
                if "relax_vol" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_vol"] = default_cal_setting[
                        "relax_vol"
                    ]
            self.cal_setting = parameter["cal_setting"]
            parameter["init_from_suffix"] = parameter.get("init_from_suffix", "00")
            self.init_from_suffix = parameter["init_from_suffix"]
        self.parameter = parameter
        self.inter_param = inter_param if inter_param is not None else {"type": "vasp"}

    def make_confs(self, path_to_work, path_to_equi, refine=False):
        path_to_work = os.path.abspath(path_to_work)
        if os.path.exists(path_to_work):
            #dlog.warning("%s already exists" % path_to_work)
            logging.warning("%s already exists" % path_to_work)
        else:
            os.makedirs(path_to_work)
        path_to_equi = os.path.abspath(path_to_equi)

        if "start_confs_path" in self.parameter and os.path.exists(
            self.parameter["start_confs_path"]
        ):
            init_path_list = glob.glob(
                os.path.join(self.parameter["start_confs_path"], "*")
            )
            struct_init_name_list = []
            for ii in init_path_list:
                struct_init_name_list.append(ii.split("/")[-1])
            struct_output_name = path_to_work.split("/")[-2]
            assert struct_output_name in struct_init_name_list
            path_to_equi = os.path.abspath(
                os.path.join(
                    self.parameter["start_confs_path"],
                    struct_output_name,
                    "relaxation",
                    "relax_task",
                )
            )

        task_list = []
        cwd = os.getcwd()

        if self.reprod:
            print("phonon reproduce starts")
            if "init_data_path" not in self.parameter:
                raise RuntimeError("please provide the initial data path to reproduce")
            init_data_path = os.path.abspath(self.parameter["init_data_path"])
            task_list = make_repro(
                self.inter_param,
                init_data_path,
                self.init_from_suffix,
                path_to_work,
                self.parameter.get("reprod_last_frame", True),
            )
            os.chdir(cwd)

        else:
            if refine:
                print("phonon refine starts")
                task_list = make_refine(
                    self.parameter["init_from_suffix"],
                    self.parameter["output_suffix"],
                    path_to_work,
                )
                os.chdir(cwd)

            else:
                if self.inter_param["type"] == "abacus":
                    CONTCAR = abacus_utils.final_stru(path_to_equi)
                    POSCAR = "STRU"
                else:
                    CONTCAR = "CONTCAR"
                    POSCAR = "POSCAR"

                equi_contcar = os.path.join(path_to_equi, CONTCAR)
                if not os.path.exists(equi_contcar):
                    raise RuntimeError("please do relaxation first")

                if self.inter_param["type"] == "abacus":
                    stru = dpdata.System(equi_contcar, fmt="stru")
                    stru.to("contcar", "CONTCAR.tmp")
                    ptypes = vasp_utils.get_poscar_types("CONTCAR.tmp")
                    ss = Structure.from_file("CONTCAR.tmp")
                    os.remove("CONTCAR.tmp")
                else:
                    ptypes = vasp_utils.get_poscar_types(equi_contcar)
                    ss = Structure.from_file(equi_contcar)
                    # gen structure

                # get user input parameter for specific structure
                st = StructureInfo(ss)
                self.structure_type = st.lattice_structure
                type_param = self.parameter.get(self.structure_type, None)
                if type_param:
                    self.primitive = type_param.get("primitive", self.primitive)
                    self.approach = type_param.get("approach", self.approach)
                    self.supercell_size = type_param.get("supercell_size", self.supercell_size)
                    self.MESH = type_param.get("MESH", self.MESH)
                    self.PRIMITIVE_AXES = type_param.get("PRIMITIVE_AXES", self.PRIMITIVE_AXES)
                    self.BAND = type_param.get("BAND", self.BAND)
                    self.BAND_POINTS = type_param.get("BAND_POINTS", self.BAND_POINTS)
                    self.BAND_CONNECTION = type_param.get("BAND_CONNECTION", self.BAND_CONNECTION)

                os.chdir(path_to_work)
                if os.path.isfile(POSCAR):
                    os.remove(POSCAR)
                if os.path.islink(POSCAR):
                    os.remove(POSCAR)
                os.symlink(os.path.relpath(equi_contcar), POSCAR)
                #           task_poscar = os.path.join(output, 'POSCAR')

                if not self.BAND:
                    raise RuntimeError('No band_path input for phonon calculation!')
                ret = ""
                ret += "ATOM_NAME ="
                for ii in ptypes:
                    ret += " %s" % ii
                ret += "\n"
                ret += "DIM = %s %s %s\n" % (
                    self.supercell_size[0],
                    self.supercell_size[1],
                    self.supercell_size[2]
                )
                if self.MESH:
                    ret += "MESH = %s %s %s\n" % (
                        self.MESH[0], self.MESH[1], self.MESH[2]
                    )
                if self.PRIMITIVE_AXES:
                    ret += "PRIMITIVE_AXES = %s\n" % self.PRIMITIVE_AXES
                ret += "BAND = %s\n" % self.BAND
                if self.BAND_POINTS:
                    ret += "BAND_POINTS = %s\n" % self.BAND_POINTS
                if self.BAND_CONNECTION:
                    ret += "BAND_CONNECTION = %s\n" % self.BAND_CONNECTION

                ret_force_read = ret + "FORCE_CONSTANTS=READ\n"

                task_list = []
                # ------------make for abacus---------------
                if self.inter_param["type"] == "abacus":
                    # make setting.conf
                    ret_sc = ""
                    ret_sc += "DIM=%s %s %s\n" % (
                        self.supercell_size[0],
                        self.supercell_size[1],
                        self.supercell_size[2]
                    )
                    ret_sc += "ATOM_NAME ="
                    for atom in ptypes:
                        ret_sc += " %s" % (atom)
                    ret_sc += "\n"
                    with open("setting.conf", "a") as fp:
                        fp.write(ret_sc)
                    # append NUMERICAL_ORBITAL to STRU after relaxation
                    orb_file = self.inter_param.get("orb_files", None)
                    abacus_utils.append_orb_file_to_stru("STRU", orb_file, prefix='pp_orb')
                    ## generate STRU-00x
                    cmd = "phonopy setting.conf --abacus -d"
                    subprocess.call(cmd, shell=True)

                    with open("band.conf", "a") as fp:
                        fp.write(ret)
                    # generate task.000*
                    stru_list = glob.glob("STRU-0*")
                    for ii in range(len(stru_list)):
                        task_path = os.path.join(path_to_work, 'task.%06d' % ii)
                        os.makedirs(task_path, exist_ok=True)
                        os.chdir(task_path)
                        task_list.append(task_path)
                        os.symlink(os.path.join(path_to_work, stru_list[ii]), 'STRU')
                        os.symlink(os.path.join(path_to_work, 'STRU'), 'STRU.ori')
                        os.symlink(os.path.join(path_to_work, 'band.conf'), 'band.conf')
                        os.symlink(os.path.join(path_to_work, 'phonopy_disp.yaml'), 'phonopy_disp.yaml')
                        try:
                            os.symlink(os.path.join(path_to_work, 'KPT'), 'KPT')
                        except:
                            pass
                    os.chdir(cwd)
                    return task_list

                # ------------make for vasp and lammps------------
                if self.primitive:
                    subprocess.call('phonopy --symmetry', shell=True)
                    subprocess.call('cp PPOSCAR POSCAR', shell=True)
                    shutil.copyfile("PPOSCAR", "POSCAR-unitcell")
                else:
                    shutil.copyfile("POSCAR", "POSCAR-unitcell")

                # make tasks
                if self.inter_param["type"] == 'vasp':
                    cmd = "phonopy -d --dim='%d %d %d' -c POSCAR" % (
                        int(self.supercell_size[0]),
                        int(self.supercell_size[1]),
                        int(self.supercell_size[2])
                    )
                    subprocess.call(cmd, shell=True)
                    # linear response method
                    if self.approach == 'linear':
                        task_path = os.path.join(path_to_work, 'task.000000')
                        os.makedirs(task_path, exist_ok=True)
                        os.chdir(task_path)
                        task_list.append(task_path)
                        os.symlink(os.path.join(path_to_work, "SPOSCAR"), "POSCAR")
                        os.symlink(os.path.join(path_to_work, "POSCAR-unitcell"), "POSCAR-unitcell")
                        with open("band.conf", "a") as fp:
                            fp.write(ret_force_read)
                    # finite displacement method
                    elif self.approach == 'displacement':
                        poscar_list = glob.glob("POSCAR-0*")
                        for ii in range(len(poscar_list)):
                            task_path = os.path.join(path_to_work, 'task.%06d' % ii)
                            os.makedirs(task_path, exist_ok=True)
                            os.chdir(task_path)
                            task_list.append(task_path)
                            os.symlink(os.path.join(path_to_work, poscar_list[ii]), 'POSCAR')
                            os.symlink(os.path.join(path_to_work, "POSCAR-unitcell"), "POSCAR-unitcell")

                        os.chdir(path_to_work)
                        with open("band.conf", "a") as fp:
                            fp.write(ret)
                        shutil.copyfile("band.conf", "task.000000/band.conf")
                        shutil.copyfile("phonopy_disp.yaml", "task.000000/phonopy_disp.yaml")

                    else:
                        raise RuntimeError(
                            f'Unsupported phonon approach input: {self.approach}. '
                            f'Please choose from "linear" and "displacement".'
                        )
                    os.chdir(cwd)
                    return task_list
                # ----------make for lammps-------------
                elif self.inter_param["type"] in LAMMPS_TYPE:
                    task_path = os.path.join(path_to_work, 'task.000000')
                    os.makedirs(task_path, exist_ok=True)
                    os.chdir(task_path)
                    task_list.append(task_path)
                    os.symlink(os.path.join(path_to_work, "POSCAR-unitcell"), POSCAR)

                    with open("band.conf", "a") as fp:
                        fp.write(ret_force_read)
                    os.chdir(cwd)
                    return task_list
                else:
                    raise RuntimeError(
                        f'Unsupported interaction type input: {self.inter_param["type"]}'
                    )

    def post_process(self, task_list):
        cwd = os.getcwd()
        inter_type = self.inter_param["type"]
        if inter_type in LAMMPS_TYPE:
            # prepare in.lammps
            for ii in task_list:
                os.chdir(ii)
                with open("in.lammps", 'r') as f1:
                    contents = f1.readlines()
                    for jj in range(len(contents)):
                        is_pair_coeff = re.search("pair_coeff", contents[jj])
                        if is_pair_coeff:
                            pair_line_id = jj
                            break
                    del contents[pair_line_id + 1:]

                with open("in.lammps", 'w') as f2:
                    for jj in range(len(contents)):
                        f2.write(contents[jj])
                # dump phonolammps command
                phonolammps_cmd = "phonolammps in.lammps -c POSCAR --dim %s %s %s " %(
                    self.supercell_size[0], self.supercell_size[1], self.supercell_size[2]
                )
                with open("run_command", 'w') as f3:
                    f3.write(phonolammps_cmd)
        elif inter_type == "vasp":
            pass
        elif inter_type == "abacus":
            pass
        os.chdir(cwd)

    def task_type(self):
        return self.parameter["type"]

    def task_param(self):
        return self.parameter

    def _compute_lower(self, output_file, all_tasks, all_res):
        cwd = Path.cwd()
        work_path = Path(output_file).parent
        output_file = os.path.abspath(output_file)
        res_data = {}
        ptr_data = os.path.dirname(output_file) + "\n"

        if not self.reprod:
            os.chdir(work_path)
            if self.inter_param["type"] == 'abacus':
                shutil.copyfile("task.000000/band.conf", "band.conf")
                shutil.copyfile("task.000000/STRU.ori", "STRU")
                shutil.copyfile("task.000000/phonopy_disp.yaml", "phonopy_disp.yaml")
                os.system('phonopy -f task.0*/OUT.ABACUS/running_scf.log')
                os.system('phonopy -f task.0*/OUT.ABACUS/running_scf.log')
                if os.path.exists("FORCE_SETS"):
                    print('FORCE_SETS is created')
                else:
                    print('FORCE_SETS can not be created')
                os.system('phonopy band.conf --abacus')
                os.system('phonopy-bandplot --gnuplot band.yaml > band.dat')

            elif self.inter_param["type"] == 'vasp':
                shutil.copyfile("task.000000/band.conf", "band.conf")
                shutil.copyfile("task.000000/POSCAR-unitcell", "POSCAR-unitcell")

                if self.approach == "linear":
                    os.chdir(all_tasks[0])
                    assert os.path.isfile('vasprun.xml'), "vasprun.xml not found"
                    os.system('phonopy --fc vasprun.xml')
                    assert os.path.isfile('FORCE_CONSTANTS'), "FORCE_CONSTANTS not created"
                    os.system('phonopy --dim="%s %s %s" -c POSCAR-unitcell band.conf' % (
                            self.supercell_size[0],
                            self.supercell_size[1],
                            self.supercell_size[2]))
                    os.system('phonopy-bandplot --gnuplot band.yaml > band.dat')
                    print('band.dat is created')
                    shutil.copyfile("band.dat", work_path/"band.dat")

                elif self.approach == "displacement":
                    shutil.copyfile("task.000000/band.conf", "band.conf")
                    shutil.copyfile("task.000000/phonopy_disp.yaml", "phonopy_disp.yaml")
                    os.system('phonopy -f task.0*/vasprun.xml')
                    if os.path.exists("FORCE_SETS"):
                        print('FORCE_SETS is created')
                    else:
                        print('FORCE_SETS can not be created')
                    os.system('phonopy --dim="%s %s %s" -c POSCAR-unitcell band.conf' % (
                        self.supercell_size[0],
                        self.supercell_size[1],
                        self.supercell_size[2]))
                    os.system('phonopy-bandplot --gnuplot band.yaml > band.dat')

            elif self.inter_param["type"] in LAMMPS_TYPE:
                os.chdir(all_tasks[0])
                assert os.path.isfile('FORCE_CONSTANTS'), "FORCE_CONSTANTS not created"
                os.system('phonopy --dim="%s %s %s" -c POSCAR band.conf' % (
                    self.supercell_size[0], self.supercell_size[1], self.supercell_size[2])
                    )
                os.system('phonopy-bandplot --gnuplot band.yaml > band.dat')
                shutil.copyfile("band.dat", work_path/"band.dat")

        else:
            if "init_data_path" not in self.parameter:
                raise RuntimeError("please provide the initial data path to reproduce")
            init_data_path = os.path.abspath(self.parameter["init_data_path"])
            res_data, ptr_data = post_repro(
                init_data_path,
                self.parameter["init_from_suffix"],
                all_tasks,
                ptr_data,
                self.parameter.get("reprod_last_frame", True),
            )

        os.chdir(work_path)
        with open('band.dat', 'r') as f:
            ptr_data = f.read()

        result_points = ptr_data.split('\n')[1][4:]
        result_lines = ptr_data.split('\n')[2:]
        res_data[result_points] = result_lines

        with open(output_file, "w") as fp:
            json.dump(res_data, fp, indent=4)

        os.chdir(cwd)
        return res_data, ptr_data



import glob
import os
import shutil
import logging
from monty.serialization import dumpfn
from apex.core.calculator.lib import abacus_utils
from apex.core.lib import crys
from apex.core.calculator.calculator import make_calculator
from apex.core.lib.utils import create_path
from apex.core.lib.dispatcher import make_submission
from apex.core.mpdb import get_structure
from dflow.python import upload_packages
upload_packages.append(__file__)
lammps_task_type = ["deepmd", "meam", "eam_fs", "eam_alloy"]


def make_equi(confs, inter_param, relax_param):
    # find all POSCARs and their name like mp-xxx
    # ...
    logging.debug("debug info make equi")
    if "type_map" in inter_param:
        ele_list = [key for key in inter_param["type_map"].keys()]
    else:
        ele_list = [key for key in inter_param["potcars"].keys()]
    # ele_list = inter_param['type_map']
    #dlog.debug("ele_list %s" % ":".join(ele_list))
    logging.debug("ele_list %s" % ":".join(ele_list))
    conf_dirs = []
    for conf in confs:
        conf_dirs.extend(glob.glob(conf))
    conf_dirs.sort()

    # generate a list of task names like mp-xxx/relaxation/relax_task
    # ...
    cwd = os.getcwd()
    # generate poscar for single element crystal
    if len(ele_list) == 1 or "single" in inter_param:
        if "single" in inter_param:
            element_label = int(inter_param["single"])
        else:
            element_label = 0
        for ii in conf_dirs:
            os.chdir(ii)
            crys_type = ii.split("/")[-1]
            #dlog.debug("crys_type: %s" % crys_type)
            logging.debug("crys_type: %s" % crys_type)
            #dlog.debug("pwd: %s" % os.getcwd())
            logging.debug("pwd: %s" % os.getcwd())
            if crys_type == "std-fcc":
                if not os.path.exists("POSCAR"):
                    crys.fcc1(ele_list[element_label]).to("POSCAR", "POSCAR")
            elif crys_type == "std-hcp":
                if not os.path.exists("POSCAR"):
                    crys.hcp(ele_list[element_label]).to("POSCAR", "POSCAR")
            elif crys_type == "std-dhcp":
                if not os.path.exists("POSCAR"):
                    crys.dhcp(ele_list[element_label]).to("POSCAR", "POSCAR")
            elif crys_type == "std-bcc":
                if not os.path.exists("POSCAR"):
                    crys.bcc(ele_list[element_label]).to("POSCAR", "POSCAR")
            elif crys_type == "std-diamond":
                if not os.path.exists("POSCAR"):
                    crys.diamond(ele_list[element_label]).to("POSCAR", "POSCAR")
            elif crys_type == "std-sc":
                if not os.path.exists("POSCAR"):
                    crys.sc(ele_list[element_label]).to("POSCAR", "POSCAR")

            if inter_param["type"] == "abacus" and not os.path.exists("STRU"):
                abacus_utils.poscar2stru("POSCAR", inter_param, "STRU")
                os.remove("POSCAR")

            os.chdir(cwd)
    task_dirs = []
    # make task directories like mp-xxx/relaxation/relax_task
    # if mp-xxx/exists then print a warning and exit.
    # ...
    for ii in conf_dirs:
        crys_type = ii.split("/")[-1]
        logging.debug("crys_type: %s" % crys_type)

        if "mp-" in crys_type and not os.path.exists(os.path.join(ii, "POSCAR")):
            get_structure(crys_type).to("POSCAR", os.path.join(ii, "POSCAR"))
            if inter_param["type"] == "abacus" and not os.path.exists("STRU"):
                abacus_utils.poscar2stru(
                    os.path.join(ii, "POSCAR"), inter_param, os.path.join(ii, "STRU")
                )
                os.remove(os.path.join(ii, "POSCAR"))

        poscar = os.path.abspath(os.path.join(ii, "POSCAR"))
        POSCAR = "POSCAR"
        if inter_param["type"] == "abacus":
            shutil.copyfile(os.path.join(ii, "STRU"), os.path.join(ii, "STRU.bk"))
            abacus_utils.modify_stru_path(os.path.join(ii, "STRU"), "pp_orb/")
            poscar = os.path.abspath(os.path.join(ii, "STRU"))
            POSCAR = "STRU"
        if not os.path.exists(poscar):
            raise FileNotFoundError("no configuration for autotest")
        if os.path.exists(os.path.join(ii, "relaxation", "jr.json")):
            os.remove(os.path.join(ii, "relaxation", "jr.json"))

        relax_dirs = os.path.abspath(
            os.path.join(ii, "relaxation", "relax_task")
        )  # to be consistent with apex in make dispatcher
        create_path(relax_dirs)
        task_dirs.append(relax_dirs)
        os.chdir(relax_dirs)
        # copy POSCARs to mp-xxx/relaxation/relax_task
        # ...
        if os.path.isfile(POSCAR):
            os.remove(POSCAR)
        os.symlink(os.path.relpath(poscar), POSCAR)
        os.chdir(cwd)
    task_dirs.sort()
    # generate task files
    relax_param["cal_type"] = "relaxation"
    if "cal_setting" not in relax_param:
        relax_param["cal_setting"] = {
            "relax_pos": True,
            "relax_shape": True,
            "relax_vol": True,
        }
    else:
        if "relax_pos" not in relax_param["cal_setting"]:
            relax_param["cal_setting"]["relax_pos"] = True
        if "relax_shape" not in relax_param["cal_setting"]:
            relax_param["cal_setting"]["relax_shape"] = True
        if "relax_vol" not in relax_param["cal_setting"]:
            relax_param["cal_setting"]["relax_vol"] = True

    for ii in task_dirs:
        poscar = os.path.join(ii, "POSCAR")
        logging.debug("task_dir %s" % ii)
        inter = make_calculator(inter_param, poscar)
        inter.make_potential_files(ii)
        inter.make_input_file(ii, "relaxation", relax_param)


def run_equi(confs, inter_param, mdata):
    # find all POSCARs and their name like mp-xxx
    # ...
    conf_dirs = []
    for conf in confs:
        conf_dirs.extend(glob.glob(conf))
    conf_dirs.sort()

    processes = len(conf_dirs)

    # generate a list of task names like mp-xxx/relaxation/relax_task
    # ...
    work_path_list = []
    for ii in conf_dirs:
        work_path_list.append(os.path.join(ii, "relaxation"))
    all_task = []
    for ii in work_path_list:
        all_task.append(os.path.join(ii, "relax_task"))
    run_tasks = all_task

    # dispatch the tasks
    # POSCAR here is useless
    virtual_calculator = make_calculator(inter_param, "POSCAR")
    forward_files = virtual_calculator.forward_files()
    forward_common_files = virtual_calculator.forward_common_files()
    backward_files = virtual_calculator.backward_files()
    #    backward_files += logs
    machine = mdata.get("machine", None)
    resources = mdata.get("resources", None)
    command = mdata.get("run_command", None)
    group_size = mdata.get("group_size", 1)
    work_path = os.getcwd()
    print("%s --> Runing... " % (work_path))

    submission = make_submission(
        mdata_machine=machine,
        mdata_resources=resources,
        commands=[command],
        work_path=work_path,
        run_tasks=run_tasks,
        group_size=group_size,
        forward_common_files=forward_common_files,
        forward_files=forward_files,
        backward_files=backward_files,
        outlog="outlog",
        errlog="errlog",
    )
    submission.run_submission()


def post_equi(confs, inter_param):
    # find all POSCARs and their name like mp-xxx
    # ...
    conf_dirs = []
    for conf in confs:
        conf_dirs.extend(glob.glob(conf))
    conf_dirs.sort()
    task_dirs = []
    for ii in conf_dirs:
        task_dirs.append(os.path.abspath(os.path.join(ii, "relaxation", "relax_task")))
    task_dirs.sort()

    # generate a list of task names like mp-xxx/relaxation
    # ...

    # dump the relaxation result.
    for ii in task_dirs:
        poscar = os.path.join(ii, "POSCAR")
        inter = make_calculator(inter_param, poscar)
        res = inter.compute(ii)
        dumpfn(res, os.path.join(ii, "result.json"), indent=4)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="apex-flow",
    version="1.1.0",
    author="Zhuoyuan Li, Tongqi Wen",
    author_email="zhuoyli@outlook.com",
    description="Alloy Properties EXplorer using simulations",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/deepmodeling/APEX.git",
    packages=setuptools.find_packages(),
    install_requires=[
        "pydflow>=1.7.83",
        "pymatgen>=2023.8.10",
        'pymatgen-analysis-defects>=2023.8.22',
        "dpdata>=0.2.13",
        "dpdispatcher",
        "phonoLAMMPS",
        "phonopy",
        "matplotlib",
        "seekpath",
        "fpop>=0.0.7",
        "boto3"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    entry_points={'console_scripts': [
         'apex = apex.__main__:main',
     ]}
)



#!/usr/bin/python3
import os
import warnings

import numpy as np
from pymatgen.io.vasp import Incar, Kpoints, Potcar

from apex.core.lib import util
#from apex.calculator.lib.vasp import incar_upper
from dflow.python import upload_packages
upload_packages.append(__file__)


class OutcarItemError(Exception):
    pass


# def get_poscar(conf_dir) :
#     conf_path = os.path.abspath(conf_dir)
#     poscar_out = os.path.join(conf_path, 'POSCAR')
#     contcar = os.path.join(conf_path, 'POSCAR')
#     lmp_dump = glob.glob(os.path.join(conf_path, 'dump.*'))
#     lmp_dump.sort()
#     if os.path.isfile(poscar_out) :
#         pass
#     elif os.path.isfile(contcar) :
#         os.symlink(contcar, poscar_out)
#     elif len(lmp_dump) == 1 :
#         dump_file = lmp_dump[0]
#         lammps.poscar_from_last_dump(dump_file, task_poscar, deepmd_type_map)


def incar_upper(dincar):
    standard_incar = {}
    for key, val in dincar.items():
        standard_incar[key.upper()] = val
    return Incar(standard_incar)


def regulate_poscar(poscar_in, poscar_out):
    with open(poscar_in, "r") as fp:
        lines = fp.read().split("\n")
    names = lines[5].split()
    counts = [int(ii) for ii in lines[6].split()]
    assert len(names) == len(counts)
    uniq_name = []
    for ii in names:
        if not (ii in uniq_name):
            uniq_name.append(ii)
    uniq_count = np.zeros(len(uniq_name), dtype=int)
    for nn, cc in zip(names, counts):
        uniq_count[uniq_name.index(nn)] += cc
    natoms = np.sum(uniq_count)
    posis = lines[8 : 8 + natoms]
    all_lines = []
    for ele in uniq_name:
        ele_lines = []
        for ii in posis:
            ele_name = ii.split()[-1]
            if ele_name == ele:
                ele_lines.append(ii)
        all_lines += ele_lines
    all_lines.append("")
    ret = lines[0:5]
    ret.append(" ".join(uniq_name))
    ret.append(" ".join([str(ii) for ii in uniq_count]))
    ret.append("Direct")
    ret += all_lines
    with open(poscar_out, "w") as fp:
        fp.write("\n".join(ret))


def sort_poscar(poscar_in, poscar_out, new_names):
    with open(poscar_in, "r") as fp:
        lines = fp.read().split("\n")
    names = lines[5].split()
    counts = [int(ii) for ii in lines[6].split()]
    new_counts = np.zeros(len(counts), dtype=int)
    for nn, cc in zip(names, counts):
        new_counts[new_names.index(nn)] += cc
    natoms = np.sum(new_counts)
    posis = lines[8 : 8 + natoms]
    all_lines = []
    for ele in new_names:
        ele_lines = []
        for ii in posis:
            ele_name = ii.split()[-1]
            if ele_name == ele:
                ele_lines.append(ii)
        all_lines += ele_lines
    all_lines.append("")
    ret = lines[0:5]
    ret.append(" ".join(new_names))
    ret.append(" ".join([str(ii) for ii in new_counts]))
    ret.append("Direct")
    ret += all_lines
    with open(poscar_out, "w") as fp:
        fp.write("\n".join(ret))


def perturb_xz(poscar_in, poscar_out, pert=0.01):
    with open(poscar_in, "r") as fp:
        lines = fp.read().split("\n")
    zz = lines[4]
    az = [float(ii) for ii in zz.split()]
    az[0] += pert
    zz = [str(ii) for ii in az]
    zz = " ".join(zz)
    lines[4] = zz
    with open(poscar_out, "w") as fp:
        fp.write("\n".join(lines))


def reciprocal_box(box):
    rbox = np.linalg.inv(box)
    rbox = rbox.T
    # rbox = rbox / np.linalg.det(box)
    # print(np.matmul(box, rbox.T))
    # print(rbox)
    return rbox


def make_kspacing_kpoints(poscar, kspacing, kgamma):
    if type(kspacing) is not list:
        kspacing = [kspacing, kspacing, kspacing]
    with open(poscar, "r") as fp:
        lines = fp.read().split("\n")
    scale = float(lines[1])
    box = []
    for ii in range(2, 5):
        box.append([float(jj) for jj in lines[ii].split()[0:3]])
    box = np.array(box)
    box *= scale
    rbox = reciprocal_box(box)
    kpoints = [
        max(1, (np.ceil(2 * np.pi * np.linalg.norm(ii) / ks).astype(int)))
        for ii, ks in zip(rbox, kspacing)
    ]
    ret = make_vasp_kpoints(kpoints, kgamma)
    return ret


def get_energies(fname):
    if not check_finished(fname):
        warnings.warn("incomplete outcar: " + fname)
    with open(fname, "r") as fp:
        lines = fp.read().split("\n")
    try:
        ener = _get_energies(lines)
        return ener
    except OutcarItemError:
        return None


def get_boxes(fname):
    if not check_finished(fname):
        warnings.warn("incomplete outcar: " + fname)
    with open(fname, "r") as fp:
        lines = fp.read().split("\n")
    try:
        ener = _get_boxes(lines)
        return ener
    except OutcarItemError:
        return None


def get_nev(fname):
    if not check_finished(fname):
        warnings.warn("incomplete outcar: " + fname)
    with open(fname, "r") as fp:
        lines = fp.read().split("\n")
    try:
        natoms = _get_natoms(lines)
        vol = _get_volumes(lines)[-1]
        ener = _get_energies(lines)[-1]
        return natoms, ener / natoms, vol / natoms
    except OutcarItemError:
        raise OutcarItemError("cannot find the result, please check the OUTCAR")
    # print(fname, natoms, vol, ener)


def get_stress(fname):
    if not check_finished(fname):
        warnings.warn("incomplete outcar: " + fname)
    with open(fname, "r") as fp:
        lines = fp.read().split("\n")
    try:
        stress = _get_stress(lines)[-1]
        return stress
    except OutcarItemError:
        return None


def check_finished(fname):
    with open(fname, "r") as fp:
        return "Elapsed time (sec):" in fp.read()


def _get_natoms(lines):
    ipt = None
    for ii in lines:
        if "ions per type" in ii:
            ipt = [int(jj) for jj in ii.split()[4:]]
            return sum(ipt)
    raise OutcarItemError("cannot find item 'ions per type'")


def _get_energies(lines):
    items = []
    for ii in lines:
        if "free  energy   TOTEN" in ii:
            items.append(float(ii.split()[4]))
    if len(items) == 0:
        raise OutcarItemError("cannot find item 'free  energy   TOTEN'")
    return items


def _split_box_line(line):
    return [float(line[0:16]), float(line[16:29]), float(line[29:42])]


def _get_boxes(lines):
    items = []
    for idx, ii in enumerate(lines):
        tmp_box = []
        if "direct lattice vectors" in ii:
            tmp_box.append(_split_box_line(lines[idx + 1]))
            tmp_box.append(_split_box_line(lines[idx + 2]))
            tmp_box.append(_split_box_line(lines[idx + 3]))
            items.append(tmp_box)
    return np.array(items)


def _get_volumes(lines):
    items = []
    for ii in lines:
        if "volume of cell" in ii:
            items.append(float(ii.split()[4]))
    if len(items) == 0:
        raise OutcarItemError("cannot find item 'volume of cell'")
    return items


def _get_stress(lines):
    items = []
    for ii in lines:
        if "in kB" in ii:
            sv = [float(jj) for jj in ii.split()[2:8]]
            tmp = sv[4]
            sv[4] = sv[5]
            sv[5] = tmp
            items.append(util.voigt_to_stress(sv))
    if len(items) == 0:
        raise OutcarItemError("cannot find item 'in kB'")
    return items


def _compute_isif(relax_ions, relax_shape, relax_volume):
    if (relax_ions) and (not relax_shape) and (not relax_volume):
        isif = 2
    elif (relax_ions) and (relax_shape) and (relax_volume):
        isif = 3
    elif (relax_ions) and (relax_shape) and (not relax_volume):
        isif = 4
    elif (not relax_ions) and (relax_shape) and (not relax_volume):
        isif = 5
    elif (not relax_ions) and (relax_shape) and (relax_volume):
        isif = 6
    elif (not relax_ions) and (not relax_shape) and (relax_volume):
        isif = 7
    else:
        raise ValueError("unknow relax style")
    return isif


def make_vasp_static_incar(
    ecut, ediff, npar, kpar, kspacing=0.5, kgamma=True, ismear=1, sigma=0.2
):
    isif = 2
    ret = ""
    ret += "PREC=A\n"
    ret += "ENCUT=%d\n" % ecut
    ret += "# ISYM=0\n"
    ret += "ALGO=normal\n"
    ret += "EDIFF=%e\n" % ediff
    ret += "EDIFFG=-0.01\n"
    ret += "LREAL=A\n"
    ret += "NPAR=%d\n" % npar
    ret += "KPAR=%d\n" % kpar
    ret += "\n"
    ret += "ISMEAR=%d\n" % ismear
    ret += "SIGMA=%f\n" % sigma
    ret += "\n"
    ret += "ISTART=0\n"
    ret += "ICHARG=2\n"
    ret += "NELMIN=6\n"
    ret += "ISIF=%d\n" % isif
    ret += "IBRION=-1\n"
    ret += "\n"
    ret += "NSW=0\n"
    ret += "\n"
    ret += "LWAVE=F\n"
    ret += "LCHARG=F\n"
    ret += "PSTRESS=0\n"
    ret += "\n"
    if kspacing is not None:
        ret += "KSPACING=%f\n" % kspacing
    if kgamma is not None:
        if kgamma:
            ret += "KGAMMA=T\n"
        else:
            ret += "KGAMMA=F\n"
    return ret


def make_vasp_relax_incar(
    ecut,
    ediff,
    relax_ion,
    relax_shape,
    relax_volume,
    npar,
    kpar,
    kspacing=0.5,
    kgamma=True,
    ismear=1,
    sigma=0.22,
):
    isif = _compute_isif(relax_ion, relax_shape, relax_volume)
    ret = ""
    ret += "PREC=A\n"
    ret += "ENCUT=%d\n" % ecut
    ret += "# ISYM=0\n"
    ret += "ALGO=normal\n"
    ret += "EDIFF=%e\n" % ediff
    ret += "EDIFFG=-0.01\n"
    ret += "LREAL=A\n"
    ret += "NPAR=%d\n" % npar
    ret += "KPAR=%d\n" % kpar
    ret += "\n"
    ret += "ISMEAR=%d\n" % ismear
    ret += "SIGMA=%f\n" % sigma
    ret += "\n"
    ret += "ISTART=0\n"
    ret += "ICHARG=2\n"
    ret += "NELM=100\n"
    ret += "NELMIN=6\n"
    ret += "ISIF=%d\n" % isif
    ret += "IBRION=2\n"
    ret += "\n"
    ret += "NSW=50\n"
    ret += "\n"
    ret += "LWAVE=F\n"
    ret += "LCHARG=F\n"
    ret += "PSTRESS=0\n"
    ret += "\n"
    if kspacing is not None:
        ret += "KSPACING=%f\n" % kspacing
    if kgamma is not None:
        if kgamma:
            ret += "KGAMMA=T\n"
        else:
            ret += "KGAMMA=F\n"
    return ret


def make_vasp_phonon_dfpt_incar(
    ecut, ediff, npar, kpar, kspacing=0.3, kgamma=True, ismear=1, sigma=0.22
):
    isif = 2
    ret = ""
    ret += "ADDGRID=True\n"
    ret += "PREC=A\n"
    ret += "ENCUT=%d\n" % ecut
    ret += "# ISYM=0\n"
    ret += "ALGO=normal\n"
    ret += "EDIFF=%e\n" % ediff
    ret += "EDIFFG=-0.01\n"
    ret += "LREAL=A\n"
    # ret += 'NPAR=%d\n' % npar
    # ret += "KPAR=%d\n" % kpar
    ret += "\n"
    ret += "ISMEAR=%d\n" % ismear
    ret += "SIGMA=%f\n" % sigma
    ret += "\n"
    ret += "ISTART=0\n"
    ret += "ICHARG=2\n"
    ret += "NELMIN=4\n"
    ret += "ISIF=%d\n" % isif
    ret += "IBRION=8\n"
    ret += "\n"
    ret += "NSW=1\n"
    ret += "\n"
    ret += "LWAVE=F\n"
    ret += "LCHARG=F\n"
    ret += "PSTRESS=0\n"
    ret += "\n"
    if kspacing is not None:
        ret += "KSPACING=%f\n" % kspacing
    if kgamma is not None:
        if kgamma:
            ret += "KGAMMA=T\n"
        else:
            ret += "KGAMMA=F\n"
    return ret


def get_poscar_types(fname):
    with open(fname, "r") as fp:
        lines = fp.read().split("\n")
    return lines[5].split()


def get_poscar_natoms(fname):
    with open(fname, "r") as fp:
        lines = fp.read().split("\n")
    return [int(ii) for ii in lines[6].split()]


def _poscar_natoms(lines):
    numb_atoms = 0
    for ii in lines[6].split():
        numb_atoms += int(ii)
    return numb_atoms


def _poscar_scale_direct(str_in, scale):
    lines = str_in.copy()
    numb_atoms = _poscar_natoms(lines)
    pscale = float(lines[1])
    pscale = pscale * scale
    lines[1] = str(pscale) + "\n"
    return lines


def _poscar_scale_cartesian(str_in, scale):
    lines = str_in.copy()
    numb_atoms = _poscar_natoms(lines)
    # scale box
    for ii in range(2, 5):
        boxl = lines[ii].split()
        boxv = [float(ii) for ii in boxl]
        boxv = np.array(boxv) * scale
        lines[ii] = "%.16e %.16e %.16e\n" % (boxv[0], boxv[1], boxv[2])
    # scale coord
    for ii in range(8, 8 + numb_atoms):
        cl = lines[ii].split()
        cv = [float(ii) for ii in cl]
        cv = np.array(cv) * scale
        lines[ii] = "%.16e %.16e %.16e\n" % (cv[0], cv[1], cv[2])
    return lines


def poscar_natoms(poscar_in):
    with open(poscar_in, "r") as fin:
        lines = list(fin)
    return _poscar_natoms(lines)


def poscar_scale(poscar_in, poscar_out, scale):
    with open(poscar_in, "r") as fin:
        lines = list(fin)
    if "D" == lines[7][0] or "d" == lines[7][0]:
        lines = _poscar_scale_direct(lines, scale)
    elif "C" == lines[7][0] or "c" == lines[7][0]:
        lines = _poscar_scale_cartesian(lines, scale)
    else:
        raise RuntimeError("Unknow poscar coord style at line 7: %s" % lines[7])
    with open(poscar_out, "w") as fout:
        fout.write("".join(lines))


def poscar_vol(poscar_in):
    with open(poscar_in, "r") as fin:
        lines = list(fin)
    box = []
    for ii in range(2, 5):
        words = lines[ii].split()
        vec = [float(jj) for jj in words]
        box.append(vec)
    scale = float(lines[1].split()[0])
    box = np.array(box)
    box *= scale
    return np.linalg.det(box)


def _make_vasp_kp_gamma(kpoints):
    ret = ""
    ret += "Automatic mesh\n"
    ret += "0\n"
    ret += "Gamma\n"
    ret += "%d %d %d\n" % (kpoints[0], kpoints[1], kpoints[2])
    ret += "0  0  0\n"
    return ret


def _make_vasp_kp_mp(kpoints):
    ret = ""
    ret += "K-Points\n"
    ret += " 0\n"
    ret += "Monkhorst Pack\n"
    ret += "%d %d %d\n" % (kpoints[0], kpoints[1], kpoints[2])
    ret += " 0  0  0\n"
    return ret


def make_vasp_kpoints(kpoints, kgamma=False):
    if kgamma:
        ret = _make_vasp_kp_gamma(kpoints)
    else:
        ret = _make_vasp_kp_mp(kpoints)
    return ret


def make_vasp_kpoints_from_incar(work_dir, jdata):
    cwd = os.getcwd()
    fp_aniso_kspacing = jdata.get("fp_aniso_kspacing")
    os.chdir(work_dir)
    # get kspacing and kgamma from incar
    assert os.path.exists("INCAR")
    with open("INCAR") as fp:
        incar = fp.read()
    standard_incar = incar_upper(Incar.from_string(incar))
    if fp_aniso_kspacing is None:
        try:
            kspacing = standard_incar["KSPACING"]
        except KeyError:
            raise RuntimeError("KSPACING must be given in INCAR")
    else:
        kspacing = fp_aniso_kspacing
    try:
        gamma = standard_incar["KGAMMA"]
        if isinstance(gamma, bool):
            pass
        else:
            if gamma[0].upper() == "T":
                gamma = True
            else:
                gamma = False
    except KeyError:
        raise RuntimeError("KGAMMA must be given in INCAR")
    # check poscar
    assert os.path.exists("POSCAR")
    # make kpoints
    ret = make_kspacing_kpoints("POSCAR", kspacing, gamma)
    kp = Kpoints.from_string(ret)
    kp.write_file("KPOINTS")
    os.chdir(cwd)





import glob
import json
import os
import re
import logging

import dpdata
import numpy as np
from monty.serialization import dumpfn, loadfn
from pymatgen.core.structure import Structure
from pymatgen.core.surface import SlabGenerator
from pymatgen.analysis.diffraction.tem import TEMCalculator

from apex.core.calculator.lib import abacus_utils
from apex.core.calculator.lib import vasp_utils
from apex.core.property.Property import Property
from apex.core.refine import make_refine
from apex.core.reproduce import make_repro, post_repro
from apex.core.structure import StructureInfo
from apex.core.lib.slab_orientation import SlabSlipSystem
from apex.core.lib.trans_tools import trans_mat_basis
from apex.core.lib.trans_tools import (plane_miller_bravais_to_miller,
                                       direction_miller_bravais_to_miller)
from dflow.python import upload_packages

upload_packages.append(__file__)


class Gamma(Property):
    """
    Calculation of common gamma lines for bcc and fcc
    """

    def __init__(self, parameter, inter_param=None):
        parameter["reproduce"] = parameter.get("reproduce", False)
        self.reprod = parameter["reproduce"]
        if not self.reprod:
            if not ("init_from_suffix" in parameter and "output_suffix" in parameter):
                self.plane_miller = parameter.get("plane_miller", None)
                self.slip_direction = parameter.get("slip_direction", None)
                self.slip_length = parameter.get("slip_length", None)
                self.plane_shift = parameter.get("plane_shift", 0)
                parameter["supercell_size"] = parameter.get("supercell_size", (1, 1, 5))
                self.supercell_size = parameter["supercell_size"]
                parameter["vacuum_size"] = parameter.get("vacuum_size", 0)
                self.vacuum_size = parameter["vacuum_size"]
                parameter["add_fix"] = parameter.get(
                    "add_fix", ["true", "true", "false"]
                )  # standard method
                self.add_fix = parameter["add_fix"]
                parameter["n_steps"] = parameter.get("n_steps", 10)
                self.n_steps = parameter["n_steps"]
                self.atom_num = None
            parameter["cal_type"] = parameter.get("cal_type", "relaxation")
            self.cal_type = parameter["cal_type"]
            default_cal_setting = {
                "relax_pos": True,
                "relax_shape": False,
                "relax_vol": False,
            }
            if "cal_setting" not in parameter:
                parameter["cal_setting"] = default_cal_setting
            else:
                if "relax_pos" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_pos"] = default_cal_setting[
                        "relax_pos"
                    ]
                if "relax_shape" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_shape"] = default_cal_setting[
                        "relax_shape"
                    ]
                if "relax_vol" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_vol"] = default_cal_setting[
                        "relax_vol"
                    ]
            self.cal_setting = parameter["cal_setting"]
        else:
            parameter["cal_type"] = "static"
            self.cal_type = parameter["cal_type"]
            default_cal_setting = {
                "relax_pos": False,
                "relax_shape": False,
                "relax_vol": False,
            }
            if "cal_setting" not in parameter:
                parameter["cal_setting"] = default_cal_setting
            else:
                if "relax_pos" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_pos"] = default_cal_setting[
                        "relax_pos"
                    ]
                if "relax_shape" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_shape"] = default_cal_setting[
                        "relax_shape"
                    ]
                if "relax_vol" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_vol"] = default_cal_setting[
                        "relax_vol"
                    ]
            self.cal_setting = parameter["cal_setting"]
            parameter["init_from_suffix"] = parameter.get("init_from_suffix", "00")
            self.init_from_suffix = parameter["init_from_suffix"]
        self.parameter = parameter
        self.inter_param = inter_param if inter_param != None else {"type": "vasp"}

    def make_confs(self, path_to_work, path_to_equi, refine=False):
        path_to_work = os.path.abspath(path_to_work)
        if os.path.exists(path_to_work):
            logging.warning("%s already exists" % path_to_work)
        else:
            os.makedirs(path_to_work)
        path_to_equi = os.path.abspath(path_to_equi)
        if "start_confs_path" in self.parameter and os.path.exists(
                self.parameter["start_confs_path"]
        ):
            init_path_list = glob.glob(
                os.path.join(self.parameter["start_confs_path"], "*")
            )
            struct_init_name_list = []
            for ii in init_path_list:
                struct_init_name_list.append(ii.split("/")[-1])
            struct_output_name = path_to_work.split("/")[-2]
            assert struct_output_name in struct_init_name_list
            path_to_equi = os.path.abspath(
                os.path.join(
                    self.parameter["start_confs_path"],
                    struct_output_name,
                    "relaxation",
                    "relax_task",
                )
            )

        task_list = []
        cwd = os.getcwd()

        if self.reprod:
            print("gamma line reproduce starts")
            if "init_data_path" not in self.parameter:
                raise RuntimeError("please provide the initial data path to reproduce")
            init_data_path = os.path.abspath(self.parameter["init_data_path"])
            task_list = make_repro(
                init_data_path,
                self.init_from_suffix,
                path_to_work,
                self.parameter.get("reprod_last_frame", True),
            )
            os.chdir(cwd)

        else:
            if refine:
                print("gamma line refine starts")
                task_list = make_refine(
                    self.parameter["init_from_suffix"],
                    self.parameter["output_suffix"],
                    path_to_work,
                )
                os.chdir(cwd)
                # record miller
                init_from_path = re.sub(
                    self.parameter["output_suffix"][::-1],
                    self.parameter["init_from_suffix"][::-1],
                    path_to_work[::-1],
                    count=1,
                )[::-1]
                task_list_basename = list(map(os.path.basename, task_list))

                for ii in task_list_basename:
                    init_from_task = os.path.join(init_from_path, ii)
                    output_task = os.path.join(path_to_work, ii)
                    os.chdir(output_task)
                    if os.path.isfile("miller.json"):
                        os.remove("miller.json")
                    if os.path.islink("miller.json"):
                        os.remove("miller.json")
                    os.symlink(
                        os.path.relpath(os.path.join(init_from_task, "miller.json")),
                        "miller.json",
                    )
                os.chdir(cwd)

            else:
                if self.inter_param["type"] == "abacus":
                    CONTCAR = abacus_utils.final_stru(path_to_equi)
                    POSCAR = "STRU"
                else:
                    CONTCAR = "CONTCAR"
                    POSCAR = "POSCAR"

                equi_contcar = os.path.join(path_to_equi, CONTCAR)
                if not os.path.exists(equi_contcar):
                    raise RuntimeError("please do relaxation first")
                # print("we now only support gamma line calculation for BCC FCC and HCP metals")
                # print(
                #    f"supported slip systems are:\n{SlabSlipSystem.hint_string()}"
                # )

                if self.inter_param["type"] == "abacus":
                    stru = dpdata.System(equi_contcar, fmt="stru")
                    stru.to("contcar", "CONTCAR.tmp")
                    ptypes = vasp_utils.get_poscar_types("CONTCAR.tmp")
                    ss = Structure.from_file("CONTCAR.tmp")
                    os.remove("CONTCAR.tmp")
                else:
                    ptypes = vasp_utils.get_poscar_types(equi_contcar)
                    # read structure from relaxed CONTCAR
                    ss = Structure.from_file(equi_contcar)

                # rewrite new CONTCAR with direct coords
                os.chdir(path_to_equi)
                ss.to("CONTCAR.direct", "POSCAR")
                # re-read new CONTCAR
                ss = Structure.from_file("CONTCAR.direct")
                # get structure type
                st = StructureInfo(ss)
                self.structure_type = st.lattice_structure
                self.conv_std_structure = st.conventional_structure
                relax_a = self.conv_std_structure.lattice.a
                relax_b = self.conv_std_structure.lattice.b
                relax_c = self.conv_std_structure.lattice.c
                # get user input slip parameter for specific structure
                type_param = self.parameter.get(self.structure_type, None)
                if type_param:
                    self.plane_miller = type_param.get("plane_miller", self.plane_miller)
                    self.slip_direction = type_param.get("slip_direction", self.slip_direction)
                    self.slip_length = type_param.get("slip_length", self.slip_length)
                    self.plane_shift = type_param.get("plane_shift", self.plane_shift)
                    self.supercell_size = type_param.get("supercell_size", self.supercell_size)
                    self.vacuum_size = type_param.get("vacuum_size", self.vacuum_size)
                    self.add_fix = type_param.get("add_fix", self.add_fix)
                    self.n_steps = type_param.get("n_steps", self.n_steps)
                if not (self.plane_miller and self.slip_direction):
                    raise RuntimeError(f'fail to obtain both slip plane '
                                       f'and slip direction info from input json file!')

                if not self.structure_type in ['bcc', 'fcc', 'hcp']:
                    logging.warning(
                        f'Gamma line function for {self.structure_type} '
                        f'is not fully supported so far.\n'
                        f'You may need to double check the generated slab structures'
                    )
                # gen initial slab
                (plane_miller, slip_direction,
                 slip_length, Q) = self.__convert_input_miller(self.conv_std_structure)
                slab = self.__gen_slab_pmg(self.conv_std_structure, plane_miller, trans_matrix=Q)
                self.atom_num = len(slab.sites)

                os.chdir(path_to_work)
                if os.path.isfile(POSCAR):
                    os.remove(POSCAR)
                if os.path.islink(POSCAR):
                    os.remove(POSCAR)
                os.symlink(os.path.relpath(equi_contcar), POSCAR)
                # task_poscar = os.path.join(output, 'POSCAR')
                count = 0
                # define slip vector
                if type(slip_length) == int or type(slip_length) == float:
                    frac_slip_vec = np.array([slip_length, 0, 0]) * relax_a
                else:
                    # for Sequence[int|float, int|float, int|float] type
                    try:
                        slip_vector_cartesian = np.multiply(np.array(slip_length),
                                                            np.array([relax_a, relax_b, relax_c]))
                        norm_length = np.linalg.norm(slip_vector_cartesian, 2)
                        frac_slip_vec = np.array([norm_length, 0, 0])
                    except:
                        raise RuntimeError(
                            'Only int | float or '
                            'Sequence[int | float, int | float, int | float] is allowed for the input_length'
                        )
                # get displaced structure
                for obtained_slab in self.__displace_slab_generator(slab,
                                                                    disp_vector=frac_slip_vec,
                                                                    is_frac=False):
                    output_task = os.path.join(path_to_work, "task.%06d" % count)
                    os.makedirs(output_task, exist_ok=True)
                    os.chdir(output_task)
                    for jj in ["INCAR", "POTCAR", POSCAR, "conf.lmp", "in.lammps"]:
                        if os.path.exists(jj):
                            os.remove(jj)
                    task_list.append(output_task)
                    # print("# %03d generate " % ii, output_task)
                    print(
                        "# %03d generate " % count,
                        output_task,
                        " \t %d atoms" % len(obtained_slab.sites)
                    )
                    # make confs
                    obtained_slab.to("POSCAR.tmp", "POSCAR")
                    vasp_utils.regulate_poscar("POSCAR.tmp", "POSCAR")
                    vasp_utils.sort_poscar("POSCAR", "POSCAR", ptypes)
                    if self.inter_param["type"] == "abacus":
                        abacus_utils.poscar2stru("POSCAR", self.inter_param, "STRU")
                        os.remove("POSCAR")
                    # vasp.perturb_xz('POSCAR', 'POSCAR', self.pert_xz)
                    # record miller
                    dumpfn(self.plane_miller, "miller.json")
                    count += 1
                os.chdir(cwd)

        return task_list

    def __convert_input_miller(self, structure: Structure):
        plane_miller = tuple(self.plane_miller)
        slip_direction = tuple(self.slip_direction)
        slip_length = self.slip_length
        # get search key string
        plane_str = ''.join([str(i) for i in plane_miller])
        slip_str = ''.join([str(i) for i in slip_direction])
        combined_key = 'x'.join([plane_str, slip_str])
        l2_normalize_1d = lambda v: v / np.linalg.norm(v, 2)
        # try to get default slip system from pre-defined dict
        dir_dict = SlabSlipSystem.atomic_system_dict()
        try:
            system = dir_dict[self.structure_type]
            (plane_miller, x_miller,
             xy_miller, stored_slip_length) = system[combined_key].values()
        except KeyError:
            logging.warning(
                'Warning:\n'
                'The input slip system is not pre-defined in the Gamma module!\n'
                'We highly recommend you to double check the slab structure generated'
                'of an undefined slip system, as it may not be what you expected, '
                'especially for a HCP structure.'
            )
            x_miller = slip_direction
            if not slip_length:
                slip_length = 1
            if self.structure_type == 'hcp' and (len(self.plane_miller) == 4 or len(self.slip_direction) == 4):
                if len(plane_miller) == 4:
                    plane_miller = plane_miller_bravais_to_miller(self.plane_miller)
                if len(x_miller) == 4:
                    x_miller = direction_miller_bravais_to_miller(self.slip_direction)
            # check user input miller index
            dir_dot = np.array(plane_miller).dot(np.array(x_miller))
            if not dir_dot == 0:
                raise RuntimeError(f'slip direction {self.slip_direction} is not '
                                   f'on plane given {self.plane_miller}')
            # Express miller_index in the conventional standard cartesian coordinate system
            x_cartesian = np.dot(np.array(x_miller), structure.lattice.matrix)
            z_cartesian = np.dot(np.array(plane_miller), structure.lattice.matrix)
            x_cartesian_unit_vector = l2_normalize_1d(x_cartesian)
            y_cartesian_unit_vector = l2_normalize_1d(np.cross(z_cartesian, x_cartesian))
            z_cartesian_unit_vector = l2_normalize_1d(np.cross(x_cartesian_unit_vector,
                                                               y_cartesian_unit_vector))
        else:
            if not slip_length:
                slip_length = stored_slip_length
            x_cartesian = np.dot(np.array(x_miller), structure.lattice.matrix)
            xy_cartesian = np.dot(np.array(xy_miller), structure.lattice.matrix)
            x_cartesian_unit_vector = l2_normalize_1d(x_cartesian)
            z_cartesian_unit_vector = l2_normalize_1d(np.cross(x_cartesian, xy_cartesian))
            y_cartesian_unit_vector = l2_normalize_1d(np.cross(z_cartesian_unit_vector,
                                                                x_cartesian_unit_vector))
        finally:
            reoriented_basis = np.array([x_cartesian_unit_vector,
                                        y_cartesian_unit_vector,
                                        z_cartesian_unit_vector])
            # Transform the lattice vectors of the slab
            Q = trans_mat_basis(reoriented_basis)

        return plane_miller, x_miller, slip_length, Q

    def __gen_slab_pmg(self, structure: Structure,
                       plane_miller, trans_matrix=None) -> Structure:
        # Get slab inter-plane distance
        tem_calc_obj = TEMCalculator()
        spacing_dict = tem_calc_obj.get_interplanar_spacings(self.conv_std_structure,
                                                             [plane_miller])
        slab_size = spacing_dict[plane_miller] * self.supercell_size[2]
        # Generate slab via Pymatgen
        slabGen = SlabGenerator(structure, miller_index=plane_miller,
                                min_slab_size=slab_size, min_vacuum_size=0,
                                center_slab=True, in_unit_planes=False,
                                lll_reduce=True, reorient_lattice=False,
                                primitive=False)
        slabs_pmg = slabGen.get_slabs(ftol=0.001)
        slab = [s for s in slabs_pmg if s.miller_index == plane_miller][0]
        # If a transform matrix is passed, reorient the slab
        if trans_matrix.any():
            reoriented_lattice_vectors = [trans_matrix.dot(v) for v in slab.lattice.matrix]
            slab = Structure(lattice=np.matrix(reoriented_lattice_vectors),
                             coords=slab.frac_coords, species=slab.species)
        # Order the atoms in the lattice in the increasing order of the third lattice direction
        # n_atoms_slab = len(slab.frac_coords)
        order = zip(slab.frac_coords, slab.species)
        c_order = sorted(order, key=lambda x: x[0][2])
        sorted_frac_coords = []
        sorted_species = []
        for (frac_coord, species) in c_order:
            sorted_frac_coords.append(frac_coord)
            sorted_species.append(species)
        # add vacuum layer to the slab with height unit of angstrom
        a, b, c = slab.lattice.matrix
        slab_height = slab.lattice.matrix[2][2]
        if slab_height >= 0:
            self.is_flip = False
            elong_scale = 1 + (self.vacuum_size / slab_height)
        else:
            self.is_flip = True
            elong_scale = 1 + (-self.vacuum_size / slab_height)
        new_lattice = [a, b, elong_scale * c]
        new_frac_coords = []
        for ii in range(len(sorted_frac_coords)):
            coord = sorted_frac_coords[ii].copy()
            coord[2] = coord[2] / elong_scale
            new_frac_coords.append(coord)
        slab = Structure(lattice=np.matrix(new_lattice),
                         coords=new_frac_coords, species=sorted_species)
        # Slab area
        # slab_area = np.linalg.norm(np.cross(slab.lattice.matrix[0], slab.lattice.matrix[1]))
        # center the slab layer around the vacuum & add plane shift along z
        plane_shift_frac = self.plane_shift * structure.lattice.c / slab.lattice.matrix[2][2]
        avg_c = np.average([c[2] for c in slab.frac_coords])
        slab.translate_sites(list(range(len(slab))),
                             [0, 0, 0.5 - avg_c - plane_shift_frac])
        # replicate slab to make specific supercell
        slab.make_supercell(scaling_matrix=[self.supercell_size[0],
                                            self.supercell_size[1],
                                            1])
        return slab

    def __displace_slab_generator(self, slab: Structure,
                                  disp_vector, is_frac=True,
                                  to_unit_cell=True) -> Structure:
        # generator of displaced slab structures
        yield slab.copy()
        # return list of atoms number to be displaced which above 0.5 z
        disp_atoms_list = np.where(slab.frac_coords[:, 2] > 0.5)[0]
        for _ in list(range(self.n_steps)):
            frac_disp = 1 / self.n_steps
            unit_vector = frac_disp * np.array(disp_vector)
            slab.translate_sites(
                indices=disp_atoms_list,
                vector=unit_vector,
                frac_coords=is_frac,
                to_unit_cell=to_unit_cell,
            )
            yield slab.copy()

    def __poscar_fix(self, poscar) -> None:
        # add position fix condition of x and y in POSCAR
        insert_pos = -self.atom_num
        fix_dict = {"true": "F", "false": "T"}
        add_fix_str = (
                " "
                + fix_dict[self.add_fix[0]]
                + " "
                + fix_dict[self.add_fix[1]]
                + " "
                + fix_dict[self.add_fix[2]]
                + "\n"
        )
        with open(poscar, "r") as fin1:
            contents = fin1.readlines()
            contents.insert(insert_pos - 1, "Selective dynamics\n")
            for ii in range(insert_pos, 0, 1):
                contents[ii] = contents[ii].replace("\n", "")
                content_split = contents[ii].split(' ')
                if len(content_split[-1]) < 3:
                    content_split.pop()
                contents[ii] = ' '.join(content_split)
                contents[ii] += add_fix_str
        with open(poscar, "w") as fin2:
            for ii in range(len(contents)):
                fin2.write(contents[ii])

    def __stru_fix(self, stru) -> None:
        fix_dict = {"true": True, "false": False}
        fix_xyz = [fix_dict[i] for i in self.addfix]
        abacus.stru_fix_atom(stru, fix_atom=fix_xyz)

    def __inLammpes_fix(self, inLammps) -> None:
        # add position fix condition of x and y of in.lammps
        fix_dict = {"true": "0", "false": "NULL"}
        add_fix_str = (
                "fix             1 all setforce"
                + " "
                + fix_dict[self.add_fix[0]]
                + " "
                + fix_dict[self.add_fix[1]]
                + " "
                + fix_dict[self.add_fix[2]]
                + "\n"
        )
        with open(inLammps, "r") as fin1:
            contents = fin1.readlines()
            for ii in range(len(contents)):
                upper = re.search("variable        N equal count\(all\)", contents[ii])
                lower = re.search("min_style       cg", contents[ii])
                if lower:
                    lower_id = ii
                    # print(lower_id)
                elif upper:
                    upper_id = ii
                    # print(upper_id)
            del contents[lower_id + 1:upper_id - 1]
            contents.insert(lower_id + 1, add_fix_str)
        with open(inLammps, "w") as fin2:
            for ii in range(len(contents)):
                fin2.write(contents[ii])

    def post_process(self, task_list):
        if self.add_fix:
            count = 0
            for ii in task_list:
                count += 1
                inter = os.path.join(ii, "inter.json")
                poscar = os.path.join(ii, "POSCAR")
                calc_type = loadfn(inter)["type"]
                if calc_type == "vasp":
                    self.__poscar_fix(poscar)
                elif calc_type == "abacus":
                    self.__stru_fix(os.path.join(ii, "STRU"))
                else:
                    inLammps = os.path.join(ii, "in.lammps")
                    if count == 1:
                        self.__inLammpes_fix(inLammps)

    def task_type(self):
        return self.parameter["type"]

    def task_param(self):
        return self.parameter

    def _compute_lower(self, output_file, all_tasks, all_res):
        output_file = os.path.abspath(output_file)
        res_data = {}
        ptr_data = os.path.dirname(output_file) + "\n"

        if not self.reprod:
            """
            ptr_data += (
                str(tuple(self.miller_index))
                + " plane along "
                + str(self.displace_direction)
            )
            """
            ptr_data += "No_task: \tDisplacement \tStacking_Fault_E(J/m^2) EpA(eV) slab_equi_EpA(eV)\n"
            all_tasks.sort()
            task_result_slab_equi = loadfn(
                os.path.join(all_tasks[0], "result_task.json")
            )
            for ii in all_tasks:
                task_result = loadfn(os.path.join(ii, "result_task.json"))
                natoms = np.sum(task_result["atom_numbs"])
                epa = task_result["energies"][-1] / natoms
                equi_epa_slab = task_result_slab_equi["energies"][-1] / natoms
                AA = np.linalg.norm(
                    np.cross(task_result["cells"][0][0], task_result["cells"][0][1])
                )

                equi_path = os.path.abspath(
                    os.path.join(
                        os.path.dirname(output_file), "../relaxation/relax_task"
                    )
                )
                equi_result = loadfn(os.path.join(equi_path, "result.json"))
                equi_epa = equi_result["energies"][-1] / np.sum(
                    equi_result["atom_numbs"]
                )
                structure_dir = os.path.basename(ii)

                Cf = 1.60217657e-16 / 1e-20 * 0.001
                sfe = (
                        (
                                task_result["energies"][-1]
                                - task_result_slab_equi["energies"][-1]
                        )
                        / AA
                        * Cf
                )

                miller_index = loadfn(os.path.join(ii, "miller.json"))
                ptr_data += "%-25s     %7.2f   %7.3f    %8.3f %8.3f\n" % (
                    str(miller_index) + "-" + structure_dir + ":",
                    int(ii[-4:]) / self.n_steps,
                    sfe,
                    epa,
                    equi_epa_slab,
                )
                res_data[int(ii[-4:]) / self.n_steps] = [sfe, epa, equi_epa]

        else:
            if "init_data_path" not in self.parameter:
                raise RuntimeError("please provide the initial data path to reproduce")
            init_data_path = os.path.abspath(self.parameter["init_data_path"])
            res_data, ptr_data = post_repro(
                init_data_path,
                self.parameter["init_from_suffix"],
                all_tasks,
                ptr_data,
                self.parameter.get("reprod_last_frame", True),
            )

        with open(output_file, "w") as fp:
            json.dump(res_data, fp, indent=4)

        return res_data, ptr_data



import os
import time
from typing import (
    Optional,
    Type,
    Union,
    List
)
from monty.serialization import loadfn
import dflow
from dflow import (
    Step,
    upload_artifact,
    download_artifact,
    Workflow
)
from dflow.python.op import OP
from dflow.plugins.dispatcher import DispatcherExecutor
from apex.superop.RelaxationFlow import RelaxationFlow
from apex.superop.PropertyFlow import PropertyFlow
from apex.op.relaxation_ops import RelaxMake, RelaxPost
from apex.op.property_ops import PropsMake, PropsPost

from dflow.python import upload_packages
upload_packages.append(__file__)


def json2dict(function):
    def wrapper(*args, **kwargs):
        # check input parameter and try to convert to dict if is json file
        for k, v in kwargs.items():
            if isinstance(v, os.PathLike) or isinstance(v, str):
                try:
                    kwargs[k] = loadfn(v)
                except Exception:
                    pass
        function(*args, **kwargs)
    return wrapper


class FlowFactory:
    def __init__(
        self,
        make_image: str,
        run_image: str,
        post_image: str,
        run_command: str,
        calculator: str,
        run_op: Type[OP],
        relax_make_op: Type[OP] = RelaxMake,
        relax_post_op: Type[OP] = RelaxPost,
        props_make_op: Type[OP] = PropsMake,
        props_post_op: Type[OP] = PropsPost,
        group_size: Optional[int] = None,
        pool_size: Optional[int] = None,
        executor: Optional[DispatcherExecutor] = None,
        upload_python_packages: Optional[List[os.PathLike]] = None,
    ):
        self.relax_make_op = relax_make_op
        self.relax_post_op = relax_post_op
        self.props_make_op = props_make_op
        self.props_post_op = props_post_op
        self.run_op = run_op
        self.make_image = make_image
        self.run_image = run_image
        self.post_image = post_image
        self.run_command = run_command
        self.calculator = calculator
        self.group_size = group_size
        self.pool_size = pool_size
        self.executor = executor
        self.upload_python_packages = upload_python_packages

    @staticmethod
    def download(
            wf,
            step_name: str,
            artifacts_key: str,
            work_dir: Union[os.PathLike, str] = '.'
    ):
        while wf.query_status() in ["Pending", "Running"]:
            time.sleep(4)
        assert (wf.query_status() == 'Succeeded')
        print(f'Workflow finished (ID: {wf.id}, UID: {wf.uid})')
        print('Retrieving finished tasks to local...')
        step = wf.query_step(name=step_name)[0]
        download_artifact(
            step.outputs.artifacts[artifacts_key],
            path=work_dir
        )

    def _set_relax_flow(
            self,
            input_work_dir: dflow.common.S3Artifact,
            relax_parameter: dict
    ) -> Step:
        relaxation_flow = RelaxationFlow(
            name='relaxation-flow',
            make_op=self.relax_make_op,
            run_op=self.run_op,
            post_op=self.relax_post_op,
            make_image=self.make_image,
            run_image=self.run_image,
            post_image=self.post_image,
            run_command=self.run_command,
            calculator=self.calculator,
            group_size=self.group_size,
            pool_size=self.pool_size,
            executor=self.executor,
            upload_python_packages=self.upload_python_packages
        )
        relaxation = Step(
            name='relaxation-cal',
            template=relaxation_flow,
            artifacts={
                "input_work_path": input_work_dir
            },
            parameters={
                "flow_id": "relaxflow",
                "parameter": relax_parameter
            },
            key="relaxationcal"
        )
        return relaxation

    def _set_props_flow(
            self,
            input_work_dir: dflow.common.S3Artifact,
            props_parameter: dict
    ) -> Step:
        property_flow = PropertyFlow(
            name='property-flow',
            make_op=self.props_make_op,
            run_op=self.run_op,
            post_op=self.props_post_op,
            make_image=self.make_image,
            run_image=self.run_image,
            post_image=self.post_image,
            run_command=self.run_command,
            calculator=self.calculator,
            group_size=self.group_size,
            pool_size=self.pool_size,
            executor=self.executor,
            upload_python_packages=self.upload_python_packages
        )
        property = Step(
            name='property-cal',
            template=property_flow,
            artifacts={
                "input_work_path": input_work_dir
            },
            parameters={
                "flow_id": "propertyflow",
                "parameter": props_parameter
            },
            key="propertycal"
        )
        return property

    @json2dict
    def submit_relax(
            self,
            work_dir: Union[os.PathLike, str],
            relax_parameter: dict,
            labels: Optional[dict] = None
    ):
        wf = Workflow(name='relaxation', labels=labels)
        relaxation = self._set_relax_flow(
            input_work_dir=upload_artifact(work_dir),
            relax_parameter=relax_parameter
        )
        wf.add(relaxation)
        wf.submit()
        self.download(
            wf, step_name='relaxation-cal',
            artifacts_key='retrieve_path',
            work_dir=work_dir
        )

    @json2dict
    def submit_props(
            self,
            work_dir: Union[os.PathLike, str],
            props_parameter: dict,
            labels: Optional[dict] = None
    ):
        wf = Workflow(name='property', labels=labels)
        properties = self._set_props_flow(
            input_work_dir=upload_artifact(work_dir),
            props_parameter=props_parameter
        )
        wf.add(properties)
        wf.submit()
        self.download(
            wf, step_name='property-cal',
            artifacts_key='retrieve_path',
            work_dir=work_dir
        )

    @json2dict
    def submit_joint(
            self,
            work_dir: Union[os.PathLike, str],
            relax_parameter: dict,
            props_parameter: dict,
            labels: Optional[dict] = None
    ):
        wf = Workflow(name='joint', labels=labels)
        relaxation = self._set_relax_flow(
            input_work_dir=upload_artifact(work_dir),
            relax_parameter=relax_parameter
        )
        properties = self._set_props_flow(
            input_work_dir=relaxation.outputs.artifacts["output_all"],
            props_parameter=props_parameter
        )
        wf.add(relaxation)
        wf.add(properties)
        wf.submit()
        self.download(
            wf, step_name='property-cal',
            artifacts_key='retrieve_path',
            work_dir=work_dir
        )




import os
import logging

from dpdata import LabeledSystem
from monty.serialization import dumpfn
from pymatgen.core.structure import Structure
from pymatgen.io.vasp import Incar, Kpoints

#from dpgen import dlog
from apex.core.calculator.Task import Task
from apex.core.calculator.lib import vasp_utils
from apex.core.calculator.lib.vasp_utils import incar_upper
from apex.utils import sepline
from dflow.python import upload_packages
upload_packages.append(__file__)


class VASP(Task):
    def __init__(self, inter_parameter, path_to_poscar):
        self.inter = inter_parameter
        self.inter_type = inter_parameter["type"]
        self.incar = inter_parameter["incar"]
        self.potcar_prefix = inter_parameter.get("potcar_prefix", "")
        self.potcars = inter_parameter["potcars"]
        self.path_to_poscar = path_to_poscar

    def make_potential_files(self, output_dir):
        potcar_not_link_list = ["vacancy", "interstitial"]
        task_type = output_dir.split("/")[-2].split("_")[0]

        ele_pot_list = [key for key in self.potcars.keys()]
        poscar = os.path.abspath(os.path.join(output_dir, "POSCAR"))
        pos_str = Structure.from_file(poscar)
        ele_pos_list_tmp = list(ii.as_dict()["element"] for ii in pos_str.species)

        ele_pos_list = [ele_pos_list_tmp[0]]
        for ii in range(1, len(ele_pos_list_tmp)):
            if not ele_pos_list_tmp[ii] == ele_pos_list_tmp[ii - 1]:
                ele_pos_list.append(ele_pos_list_tmp[ii])

        if task_type in potcar_not_link_list:
            with open(os.path.join(output_dir, "POTCAR"), "w") as fp:
                for ii in ele_pos_list:
                    for jj in ele_pot_list:
                        if ii == jj:
                            with open(
                                os.path.join(self.potcar_prefix, self.potcars[jj]), "r"
                            ) as fin:
                                for line in fin:
                                    print(line.strip("\n"), file=fp)

        else:
            if not os.path.isfile(os.path.join(output_dir, "../POTCAR")):
                with open(os.path.join(output_dir, "../POTCAR"), "w") as fp:
                    for ii in ele_pos_list:
                        for jj in ele_pot_list:
                            if ii == jj:
                                with open(
                                    os.path.join(self.potcar_prefix, self.potcars[jj]),
                                    "r",
                                ) as fin:
                                    for line in fin:
                                        print(line.strip("\n"), file=fp)
            cwd = os.getcwd()
            os.chdir(output_dir)
            if not os.path.islink("POTCAR"):
                os.symlink("../POTCAR", "POTCAR")
            elif not "../POTCAR" == os.readlink("POTCAR"):
                os.remove("POTCAR")
                os.symlink("../POTCAR", "POTCAR")
            os.chdir(cwd)

        dumpfn(self.inter, os.path.join(output_dir, "inter.json"), indent=4)

    def make_input_file(self, output_dir, task_type, task_param):
        sepline(ch=output_dir)
        dumpfn(task_param, os.path.join(output_dir, "task.json"), indent=4)

        assert os.path.exists(self.incar), "no INCAR file for relaxation"
        relax_incar_path = os.path.abspath(self.incar)
        incar_relax = incar_upper(Incar.from_file(relax_incar_path))

        # deal with relaxation
        prop_type = task_param.get("type", "relaxation")
        cal_type = task_param["cal_type"]
        cal_setting = task_param["cal_setting"]

        # user input INCAR for apex calculation
        if "input_prop" in cal_setting and os.path.isfile(cal_setting["input_prop"]):
            incar_prop = os.path.abspath(cal_setting["input_prop"])
            incar = incar_upper(Incar.from_file(incar_prop))
            logging.info(f"Will use user specified INCAR (path: {incar_prop}) for {prop_type} calculation")

        # revise INCAR based on the INCAR provided in the "interaction"
        else:
            if prop_type == "phonon":
                approach = task_param.get("approach", "linear")
                logging.info(f"No specification of INCAR for {prop_type} calculation, will auto generate")
                if approach == "linear":
                    incar = incar_upper(Incar.from_str(
                        vasp_utils.make_vasp_phonon_dfpt_incar(
                            ecut=650, ediff=0.0000001, npar=None, kpar=None
                        )))
                elif approach == "displacement":
                    incar = incar_upper(Incar.from_str(
                        vasp_utils.make_vasp_static_incar(
                            ecut=650, ediff=0.0000001, npar=8, kpar=1
                        )))
            else:
                if not prop_type == "relaxation":
                    logging.info(f"No specification of INCAR for {prop_type} calculation, will use INCAR for relaxation")
                incar = incar_relax

            if cal_type == "relaxation":
                relax_pos = cal_setting["relax_pos"]
                relax_shape = cal_setting["relax_shape"]
                relax_vol = cal_setting["relax_vol"]
                if [relax_pos, relax_shape, relax_vol] == [True, False, False]:
                    isif = 2
                elif [relax_pos, relax_shape, relax_vol] == [True, True, True]:
                    isif = 3
                elif [relax_pos, relax_shape, relax_vol] == [True, True, False]:
                    isif = 4
                elif [relax_pos, relax_shape, relax_vol] == [False, True, False]:
                    isif = 5
                elif [relax_pos, relax_shape, relax_vol] == [False, True, True]:
                    isif = 6
                elif [relax_pos, relax_shape, relax_vol] == [False, False, True]:
                    isif = 7
                elif [relax_pos, relax_shape, relax_vol] == [False, False, False]:
                    nsw = 0
                    isif = 2
                    if not ("NSW" in incar and incar.get("NSW") == nsw):
                        logging.info(
                            "%s setting NSW to %d"
                            % (self.make_input_file.__name__, nsw)
                        )
                        incar["NSW"] = nsw
                else:
                    raise RuntimeError("not supported calculation setting for VASP")

                if not ("ISIF" in incar and incar.get("ISIF") == isif):
                    logging.info(
                        "%s setting ISIF to %d" % (self.make_input_file.__name__, isif)
                    )
                    incar["ISIF"] = isif

            elif cal_type == "static":
                nsw = 0
                if not ("NSW" in incar and incar.get("NSW") == nsw):
                    logging.info(
                        "%s setting NSW to %d" % (self.make_input_file.__name__, nsw)
                    )
                    incar["NSW"] = nsw

            else:
                raise RuntimeError("not supported calculation type for VASP")

            if "ediff" in cal_setting:
                logging.info(
                    "%s setting EDIFF to %s"
                    % (self.make_input_file.__name__, cal_setting["ediff"])
                )
                incar["EDIFF"] = cal_setting["ediff"]

            if "ediffg" in cal_setting:
                logging.info(
                    "%s setting EDIFFG to %s"
                    % (self.make_input_file.__name__, cal_setting["ediffg"])
                )
                incar["EDIFFG"] = cal_setting["ediffg"]

            if "encut" in cal_setting:
                logging.info(
                    "%s setting ENCUT to %s"
                    % (self.make_input_file.__name__, cal_setting["encut"])
                )
                incar["ENCUT"] = cal_setting["encut"]

            if "kspacing" in cal_setting:
                logging.info(
                    "%s setting KSPACING to %s"
                    % (self.make_input_file.__name__, cal_setting["kspacing"])
                )
                incar["KSPACING"] = cal_setting["kspacing"]

            if "kgamma" in cal_setting:
                logging.info(
                    "%s setting KGAMMA to %s"
                    % (self.make_input_file.__name__, cal_setting["kgamma"])
                )
                incar["KGAMMA"] = cal_setting["kgamma"]

        try:
            kspacing = incar.get("KSPACING")
        except KeyError:
            raise RuntimeError("KSPACING must be given in INCAR")

        if "KGAMMA" in incar:
            kgamma = incar.get("KGAMMA")
        else:
            kgamma = False

        incar.write_file(os.path.join(output_dir, "../INCAR"))
        cwd = os.getcwd()
        os.chdir(output_dir)
        if not os.path.islink("INCAR"):
            os.symlink("../INCAR", "INCAR")
        elif not "../INCAR" == os.readlink("INCAR"):
            os.remove("INCAR")
            os.symlink("../INCAR", "INCAR")
        os.chdir(cwd)
        ret = vasp_utils.make_kspacing_kpoints(self.path_to_poscar, kspacing, kgamma)
        kp = Kpoints.from_str(ret)
        kp.write_file(os.path.join(output_dir, "KPOINTS"))

    def compute(self, output_dir):
        outcar = os.path.join(output_dir, "OUTCAR")
        if not os.path.isfile(outcar):
            logging.warning("cannot find OUTCAR in " + output_dir + " skip")
            return None
        else:
            ls = LabeledSystem(outcar)
            stress = []
            with open(outcar, "r") as fin:
                lines = fin.read().split("\n")
            for line in lines:
                if "in kB" in line:
                    stress_xx = float(line.split()[2])
                    stress_yy = float(line.split()[3])
                    stress_zz = float(line.split()[4])
                    stress_xy = float(line.split()[5])
                    stress_yz = float(line.split()[6])
                    stress_zx = float(line.split()[7])
                    stress.append([])
                    stress[-1].append([stress_xx, stress_xy, stress_zx])
                    stress[-1].append([stress_xy, stress_yy, stress_yz])
                    stress[-1].append([stress_zx, stress_yz, stress_zz])

            outcar_dict = ls.as_dict()
            outcar_dict["data"]["stress"] = {
                "@module": "numpy",
                "@class": "array",
                "dtype": "float64",
                "data": stress,
            }

            return outcar_dict

    def forward_files(self, property_type="relaxation"):
        return ["INCAR", "POSCAR", "KPOINTS", "POTCAR"]

    def forward_common_files(self, property_type="relaxation"):
        potcar_not_link_list = ["vacancy", "interstitial"]
        if property_type == "elastic":
            return ["INCAR", "KPOINTS", "POTCAR"]
        elif property_type in potcar_not_link_list:
            return ["INCAR"]
        else:
            return ["INCAR", "POTCAR"]

    def backward_files(self, property_type="relaxation"):
        return ["OUTCAR", "outlog", "CONTCAR", "OSZICAR", "XDATCAR"]




import glob
import os
import shutil
import sys
import unittest

from pymatgen.core.structure import Structure
from pymatgen.io.vasp import Incar

from apex.core.property.Gamma import Gamma

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"


class TestGamma(unittest.TestCase):
    def setUp(self):
        _jdata = {
            "structures": ["confs/std-fcc"],
            "interaction": {
                "type": "vasp",
                "incar": "vasp_input/INCAR_Mo",
                "potcar_prefix": "vasp_input",
                "potcars": {"Mo": "POTCAR_Mo"},
            },
            "properties": [
                {
                    "type": "gamma",
                    "plane_miller": [0, 0, 1],
                    "slip_direction": [1, 0, 0],
                    "hcp": {
                        "plane_miller": [0, 0, 0, 1],
                        "slip_direction": [2, -1, -1, 0],
                    },
                    "supercell_size": [1, 1, 10],
                    "vacuum_size": 10,
                    "add_fix": ["true", "true", "false"],
                    "n_steps": 10
                }

            ],
        }

        self.equi_path = "confs/hp-Mo/relaxation/relax_task"
        self.source_path = "equi/vasp"
        self.target_path = "confs/hp-Mo/gamma_00"
        self.res_data = "output/gamma_00/result.json"
        self.ptr_data = "output/gamma_00/result.out"

        if not os.path.exists(self.equi_path):
            os.makedirs(self.equi_path)
        if not os.path.exists(self.target_path):
            os.makedirs(self.target_path)

        self.confs = _jdata["structures"]
        self.inter_param = _jdata["interaction"]
        self.prop_param = _jdata["properties"]

        self.gamma = Gamma(_jdata["properties"][0])

    def tearDown(self):
        if os.path.exists(self.equi_path):
            shutil.rmtree(self.equi_path)
        if os.path.exists(self.target_path):
            shutil.rmtree(self.target_path)
        if os.path.exists(self.res_data):
            os.remove(self.res_data)
        if os.path.exists(self.ptr_data):
            os.remove(self.ptr_data)

    def test_task_type(self):
        self.assertEqual("gamma", self.gamma.task_type())

    def test_task_param(self):
        self.assertEqual(self.prop_param[0], self.gamma.task_param())

    def test_make_confs_bcc(self):
        if not os.path.exists(os.path.join(self.equi_path, "CONTCAR")):
            with self.assertRaises(RuntimeError):
                self.gamma.make_confs(self.target_path, self.equi_path)
        shutil.copy(
            os.path.join(self.source_path, "CONTCAR_Mo_bcc"),
            os.path.join(self.equi_path, "CONTCAR"),
        )
        task_list = self.gamma.make_confs(self.target_path, self.equi_path)
        dfm_dirs = glob.glob(os.path.join(self.target_path, "task.*"))
        self.assertEqual(len(dfm_dirs), self.gamma.n_steps + 1)

        incar0 = Incar.from_file(os.path.join("vasp_input", "INCAR.rlx"))
        incar0["ISIF"] = 4

        self.assertEqual(
            os.path.realpath(os.path.join(self.equi_path, "CONTCAR")),
            os.path.realpath(os.path.join(self.target_path, "POSCAR")),
        )
        ref_st = Structure.from_file(os.path.join(self.target_path, "POSCAR"))
        dfm_dirs.sort()
        for ii in dfm_dirs:
            st_file = os.path.join(ii, "POSCAR")
            self.assertTrue(os.path.isfile(st_file))
            st0 = Structure.from_file(st_file)
            st1_file = os.path.join(ii, "POSCAR.tmp")
            self.assertTrue(os.path.isfile(st1_file))
            st1 = Structure.from_file(st1_file)
            with open(st1_file, mode="r") as f:
                z_coord_str = f.readlines()[-1].split()[-2]
                z_coord = float(z_coord_str)
            self.assertTrue(z_coord <= 1)

    def test_compute_lower(self):
        cwd = os.getcwd()
        output_file = os.path.join(cwd, "output/gamma_00/result.json")
        all_tasks = glob.glob("output/gamma_00/task.*")
        all_tasks.sort()
        all_res = [os.path.join(task, "result_task.json") for task in all_tasks]

        self.gamma._compute_lower(output_file, all_tasks, all_res)

        self.assertTrue(os.path.isfile(self.res_data))



import os, subprocess
from pathlib import Path
from dflow.python import (
    OP,
    OPIO,
    OPIOSign,
    Artifact,
    upload_packages
)

upload_packages.append(__file__)


class RunLAMMPS(OP):
    """
    class for LAMMPS calculation
    """
    def __init__(self, infomode=1):
        self.infomode = infomode

    @classmethod
    def get_input_sign(cls):
        return OPIOSign({
            'input_lammps': Artifact(Path),
            'run_command': str
        })

    @classmethod
    def get_output_sign(cls):
        return OPIOSign({
            'backward_dir': Artifact(Path, sub_path=False)
        })

    @OP.exec_sign_check
    def execute(self, op_in: OPIO) -> OPIO:
        cwd = os.getcwd()
        os.chdir(op_in["input_lammps"])
        if os.path.exists("run_command"):
            with open("run_command", 'r') as f:
                cmd = f.read()
        else:
            cmd = op_in["run_command"]
        exit_code = subprocess.call(cmd, shell=True)
        if exit_code == 0:
            print("Call Lammps command successfully!")
        else:
            print("Call Lammps command failed with exit code:", exit_code)

        os.chdir(cwd)
        op_out = OPIO({
            "backward_dir": op_in["input_lammps"]
        })
        return op_out



import glob
import json
import logging
import os
import re

import numpy as np
from monty.serialization import dumpfn, loadfn

from apex.core.calculator.lib import abacus_utils
from apex.core.calculator.lib import vasp_utils
from apex.core.calculator.lib import abacus_scf
from apex.core.property.Property import Property
from apex.core.refine import make_refine
from apex.core.reproduce import make_repro, post_repro
from dflow.python import upload_packages
upload_packages.append(__file__)


class EOS(Property):
    def __init__(self, parameter, inter_param=None):
        parameter["reproduce"] = parameter.get("reproduce", False)
        self.reprod = parameter["reproduce"]
        if not self.reprod:
            if not ("init_from_suffix" in parameter and "output_suffix" in parameter):
                self.vol_start = parameter["vol_start"]
                self.vol_end = parameter["vol_end"]
                self.vol_step = parameter["vol_step"]
                parameter["vol_abs"] = parameter.get("vol_abs", False)
                self.vol_abs = parameter["vol_abs"]
            parameter["cal_type"] = parameter.get("cal_type", "relaxation")
            self.cal_type = parameter["cal_type"]
            default_cal_setting = {
                "relax_pos": True,
                "relax_shape": True,
                "relax_vol": False,
            }
            if "cal_setting" not in parameter:
                parameter["cal_setting"] = default_cal_setting
            else:
                if "relax_pos" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_pos"] = default_cal_setting[
                        "relax_pos"
                    ]
                if "relax_shape" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_shape"] = default_cal_setting[
                        "relax_shape"
                    ]
                if "relax_vol" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_vol"] = default_cal_setting[
                        "relax_vol"
                    ]
            self.cal_setting = parameter["cal_setting"]
        else:
            parameter["cal_type"] = "static"
            self.cal_type = parameter["cal_type"]
            default_cal_setting = {
                "relax_pos": False,
                "relax_shape": False,
                "relax_vol": False,
            }
            if "cal_setting" not in parameter:
                parameter["cal_setting"] = default_cal_setting
            else:
                if "relax_pos" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_pos"] = default_cal_setting[
                        "relax_pos"
                    ]
                if "relax_shape" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_shape"] = default_cal_setting[
                        "relax_shape"
                    ]
                if "relax_vol" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_vol"] = default_cal_setting[
                        "relax_vol"
                    ]
            self.cal_setting = parameter["cal_setting"]
            parameter["init_from_suffix"] = parameter.get("init_from_suffix", "00")
            self.init_from_suffix = parameter["init_from_suffix"]
        self.parameter = parameter
        self.inter_param = inter_param if inter_param != None else {"type": "vasp"}

    def make_confs(self, path_to_work, path_to_equi, refine=False):
        path_to_work = os.path.abspath(path_to_work)
        if os.path.exists(path_to_work):
            #dlog.warning("%s already exists" % path_to_work)
            logging.warning("%s already exists" % path_to_work)
        else:
            os.makedirs(path_to_work)
        path_to_equi = os.path.abspath(path_to_equi)

        if "start_confs_path" in self.parameter and os.path.exists(
            self.parameter["start_confs_path"]
        ):
            init_path_list = glob.glob(
                os.path.join(self.parameter["start_confs_path"], "*")
            )
            struct_init_name_list = []
            for ii in init_path_list:
                struct_init_name_list.append(ii.split("/")[-1])
            struct_output_name = path_to_work.split("/")[-2]
            assert struct_output_name in struct_init_name_list
            path_to_equi = os.path.abspath(
                os.path.join(
                    self.parameter["start_confs_path"],
                    struct_output_name,
                    "relaxation",
                    "relax_task",
                )
            )

        cwd = os.getcwd()
        task_list = []
        if self.reprod:
            print("eos reproduce starts")
            if "init_data_path" not in self.parameter:
                raise RuntimeError("please provide the initial data path to reproduce")
            init_data_path = os.path.abspath(self.parameter["init_data_path"])
            task_list = make_repro(
                self.inter_param,
                init_data_path,
                self.init_from_suffix,
                path_to_work,
                self.parameter.get("reprod_last_frame", True),
            )
            os.chdir(cwd)

        else:
            if refine:
                print("eos refine starts")
                task_list = make_refine(
                    self.parameter["init_from_suffix"],
                    self.parameter["output_suffix"],
                    path_to_work,
                )
                os.chdir(cwd)

                init_from_path = re.sub(
                    self.parameter["output_suffix"][::-1],
                    self.parameter["init_from_suffix"][::-1],
                    path_to_work[::-1],
                    count=1,
                )[::-1]
                task_list_basename = list(map(os.path.basename, task_list))

                for ii in task_list_basename:
                    init_from_task = os.path.join(init_from_path, ii)
                    output_task = os.path.join(path_to_work, ii)
                    os.chdir(output_task)
                    if os.path.isfile("eos.json"):
                        os.remove("eos.json")
                    if os.path.islink("eos.json"):
                        os.remove("eos.json")
                    os.symlink(
                        os.path.relpath(os.path.join(init_from_task, "eos.json")),
                        "eos.json",
                    )
                os.chdir(cwd)

            else:
                print(
                    "gen eos from "
                    + str(self.vol_start)
                    + " to "
                    + str(self.vol_end)
                    + " by every "
                    + str(self.vol_step)
                )
                if self.vol_abs:
                    logging.info("treat vol_start and vol_end as absolute volume")
                    #dlog.info("treat vol_start
                    # and vol_end as absolute volume")
                else:
                    logging.info("treat vol_start and vol_end as relative volume")
                    #dlog.info("treat vol_start and vol_end as relative volume")

                if self.inter_param["type"] == "abacus":
                    equi_contcar = os.path.join(
                        path_to_equi, abacus_utils.final_stru(path_to_equi)
                    )
                else:
                    equi_contcar = os.path.join(path_to_equi, "CONTCAR")

                if not os.path.isfile(equi_contcar):
                    raise RuntimeError(
                        "Can not find %s, please do relaxation first" % equi_contcar
                    )

                if self.inter_param["type"] == "abacus":
                    stru_data = abacus_scf.get_abacus_STRU(equi_contcar)
                    vol_to_poscar = (
                        abs(np.linalg.det(stru_data["cells"]))
                        / np.array(stru_data["atom_numbs"]).sum()
                    )
                else:
                    vol_to_poscar = vasp_utils.poscar_vol(equi_contcar) / vasp_utils.poscar_natoms(
                        equi_contcar
                    )
                self.parameter["scale2equi"] = []

                task_num = 0
                while self.vol_start + self.vol_step * task_num < self.vol_end:
                    # for vol in np.arange(int(self.vol_start * 100), int(self.vol_end * 100), int(self.vol_step * 100)):
                    # vol = vol / 100.0
                    vol = self.vol_start + task_num * self.vol_step
                    # task_num = int((vol - self.vol_start) / self.vol_step)
                    output_task = os.path.join(path_to_work, "task.%06d" % task_num)
                    os.makedirs(output_task, exist_ok=True)
                    os.chdir(output_task)
                    if self.inter_param["type"] == "abacus":
                        POSCAR = "STRU"
                        POSCAR_orig = "STRU.orig"
                        scale_func = abacus_utils.stru_scale
                    else:
                        POSCAR = "POSCAR"
                        POSCAR_orig = "POSCAR.orig"
                        scale_func = vasp_utils.poscar_scale

                    for ii in [
                        "INCAR",
                        "POTCAR",
                        POSCAR_orig,
                        POSCAR,
                        "conf.lmp",
                        "in.lammps",
                    ]:
                        if os.path.exists(ii):
                            os.remove(ii)
                    task_list.append(output_task)
                    os.symlink(os.path.relpath(equi_contcar), POSCAR_orig)
                    # scale = (vol / vol_to_poscar) ** (1. / 3.)

                    if self.vol_abs:
                        scale = (vol / vol_to_poscar) ** (1.0 / 3.0)
                        eos_params = {"volume": vol, "scale": scale}
                    else:
                        scale = vol ** (1.0 / 3.0)
                        eos_params = {"volume": vol * vol_to_poscar, "scale": scale}
                    dumpfn(eos_params, "eos.json", indent=4)
                    self.parameter["scale2equi"].append(scale)  # 06/22
                    scale_func(POSCAR_orig, POSCAR, scale)
                    task_num += 1
                os.chdir(cwd)
        return task_list

    def post_process(self, task_list):
        pass

    def task_type(self):
        return self.parameter["type"]

    def task_param(self):
        return self.parameter

    def _compute_lower(self, output_file, all_tasks, all_res):
        output_file = os.path.abspath(output_file)
        res_data = {}
        ptr_data = "conf_dir: " + os.path.dirname(output_file) + "\n"
        if not self.reprod:
            ptr_data += " VpA(A^3)  EpA(eV)\n"
            for ii in range(len(all_tasks)):
                # vol = self.vol_start + ii * self.vol_step
                vol = loadfn(os.path.join(all_tasks[ii], "eos.json"))["volume"]
                task_result = loadfn(all_res[ii])
                res_data[vol] = task_result["energies"][-1] / sum(
                    task_result["atom_numbs"]
                )
                ptr_data += "%7.3f  %8.4f \n" % (
                    vol,
                    task_result["energies"][-1] / sum(task_result["atom_numbs"]),
                )
                # res_data[vol] = all_res[ii]['energy'] / len(all_res[ii]['force'])
                # ptr_data += '%7.3f  %8.4f \n' % (vol, all_res[ii]['energy'] / len(all_res[ii]['force']))

        else:
            if "init_data_path" not in self.parameter:
                raise RuntimeError("please provide the initial data path to reproduce")
            init_data_path = os.path.abspath(self.parameter["init_data_path"])
            res_data, ptr_data = post_repro(
                init_data_path,
                self.parameter["init_from_suffix"],
                all_tasks,
                ptr_data,
                self.parameter.get("reprod_last_frame", True),
            )

        with open(output_file, "w") as fp:
            json.dump(res_data, fp, indent=4)

        return res_data, ptr_data



import pymatgen.core
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from dflow.python import upload_packages

upload_packages.append(__file__)

class StructureInfo(object):
    """Analyze structure type
    Arg:
        structure: pymatgen.core.Structure object
    """
    def __init__(self, structure: pymatgen.core.Structure):
        analyzer = SpacegroupAnalyzer(structure)
        self.__space_group_symbol = analyzer.get_space_group_symbol()
        self.__space_group_number = analyzer.get_space_group_number()
        self.__point_group_symbol = analyzer.get_point_group_symbol()
        self.__crystal_system = analyzer.get_crystal_system()
        self.__lattice_type = analyzer.get_lattice_type()
        self.__num_atoms = structure.num_sites
        self.__crystal_structure = self.__indentify_crystal()
        # standard structure
        self.orig_structure = structure
        self.primitive_structure = analyzer.find_primitive()
        self.conventional_structure = analyzer.get_conventional_standard_structure()

    def __indentify_crystal(self) -> str:
        if self.__lattice_type == 'cubic':
            if self.__num_atoms == 1 and self.__space_group_symbol == 'Pm-3m':
                structure_type = 'sc'
            elif self.__num_atoms == 2 and self.__space_group_symbol == 'Im-3m':
                structure_type = 'bcc'
            elif self.__num_atoms == 4 and self.__space_group_symbol == 'Fm-3m':
                structure_type = 'fcc'
            elif self.__num_atoms == 8 and self.__space_group_symbol == 'Fd-3m':
                structure_type = 'diamond'
            else:
                structure_type = 'other'

        elif self.__lattice_type == 'hexagonal':
            if self.__num_atoms == 2 and self.__space_group_symbol == 'P6_3/mmc':
                structure_type = 'hcp'
            elif self.__space_group_symbol == 'P6/mmm':
                structure_type = 'c32'
            else:
                structure_type = 'other'
        else:
            structure_type = 'other'

        return structure_type

    @property
    def space_group_symbol(self):
        return self.__space_group_symbol

    @property
    def space_group_number(self):
        return self.__space_group_number

    @property
    def point_group_symbol(self):
        return self.__point_group_symbol

    @property
    def crystal_system(self):
        return self.__crystal_system

    @property
    def lattice_type(self):
        return self.__lattice_type

    @property
    def num_atoms(self):
        return self.__num_atoms

    @property
    def lattice_structure(self):
        return self.__crystal_structure



import glob
import json
import os
import shutil
import sys
import unittest

import dpdata
import numpy as np
from monty.serialization import dumpfn, loadfn
from pymatgen.io.vasp import Incar

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"

from apex.core.common_prop import make_property
from apex.core.refine import make_refine

from .context import make_kspacing_kpoints, setUpModule


class TestMakeProperty(unittest.TestCase):
    jdata = {
        "structures": ["confs/std-fcc"],
        "interaction": {
            "type": "vasp",
            "incar": "vasp_input/INCAR.rlx",
            "potcar_prefix": "vasp_input",
            "potcars": {"Al": "POT_Al"},
        },
        "properties": [
            {
                "type": "eos",
                "skip": False,
                "vol_start": 0.8,
                "vol_end": 1.2,
                "vol_step": 0.01,
                "cal_setting": {
                    "relax_pos": True,
                    "relax_shape": True,
                    "relax_vol": False,
                    "overwrite_interaction": {
                        "type": "vasp",
                        "incar": "vasp_input/INCAR.rlx",
                        "potcar_prefix": "vasp_input",
                        "potcars": {"Al": "POT_Al"},
                    },
                },
            }
        ],
    }

    def tearDown(self):
        if os.path.exists("confs/std-fcc/eos_00"):
            shutil.rmtree("confs/std-fcc/eos_00")
        if os.path.exists("confs/std-fcc/eos_02"):
            shutil.rmtree("confs/std-fcc/eos_02")
        if os.path.exists("confs/std-fcc/relaxation"):
            shutil.rmtree("confs/std-fcc/relaxation")
        if os.path.isfile("info.log"):
            os.remove("info.log")

    def test_make_eos(self):

        pwd = os.getcwd()
        confs = self.jdata["structures"]
        inter_param = self.jdata["interaction"]
        property_list = self.jdata["properties"]

        target_path_0 = "confs/std-fcc/eos_00"
        target_path_2 = "confs/std-fcc/eos_02"
        equi_path = "confs/std-fcc/relaxation/relax_task"
        source_path = "equi/vasp"

        if not os.path.exists(equi_path):
            os.makedirs(equi_path)
        shutil.copy(
            os.path.join(source_path, "CONTCAR_Al_fcc"),
            os.path.join(equi_path, "CONTCAR"),
        )

        make_property(confs, inter_param, property_list)

        dfm_dirs_0 = glob.glob(os.path.join(target_path_0, "task.*"))

        for ii in dfm_dirs_0:
            self.assertTrue(os.path.isfile(os.path.join(ii, "POSCAR")))
            shutil.copy(os.path.join(ii, "POSCAR"), os.path.join(ii, "CONTCAR"))

        path_to_work = os.path.abspath(target_path_0)
        new_prop_list = [
            {
                "type": "eos",
                "init_from_suffix": "00",
                "output_suffix": "02",
                "cal_setting": {
                    "relax_pos": True,
                    "relax_shape": True,
                    "relax_vol": False,
                    "input_prop": "lammps_input/lammps_high",
                },
            }
        ]
        # ret=make_refine('00', '02', path_to_work)
        # self.assertEqual(len(ret),len(dfm_dirs_0))
        make_property(confs, inter_param, new_prop_list)
        self.assertTrue(os.path.isdir(path_to_work.replace("00", "02")))
        os.chdir(pwd)
        dfm_dirs_2 = glob.glob(os.path.join(target_path_2, "task.*"))
        self.assertEqual(len(dfm_dirs_2), len(dfm_dirs_0))
